#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Pythonic simple SOAP Client implementation
Using lxml based SoapEnvelope."""
import sys
import traceback
from threading import Lock
import socket
import time
import http.client as httplib
from lxml import etree as etree_

from .. import observableproperties
from .. import commlog
from ..compression import CompressionHandler
from . import soapenvelope
from ..httprequesthandler import HTTPReader, mkchunks

class HTTPConnection_NODELAY(httplib.HTTPConnection):
    def connect(self):
        httplib.HTTPConnection.connect(self)
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)



class HTTPSConnection_NODELAY(httplib.HTTPSConnection):
    def connect(self):
        httplib.HTTPSConnection.connect(self)
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)



class HTTPReturnCodeError(httplib.HTTPException):
    """ THis class is used to map http return codes to Python exceptions."""
    def __init__(self, status, reason, soapfault):
        """
        :param status: integer, e.g. 404
        param reason: the provided human readable text
        """
        super(HTTPReturnCodeError, self).__init__()
        self.status = status
        self.reason = reason
        self.soapfault = soapfault

    def __repr__(self):
        if self.soapfault:
            return 'HTTPReturnCodeError(status={}, reason={}'.format(self.status, self.soapfault)
        else:
            return 'HTTPReturnCodeError(status={}, reason={}'.format(self.status, self.reason)



class SoapClient(CompressionHandler):
    _usedSoapClients = 0
    SOCKET_TIMEOUT = 10 if sys.gettrace() is None else 1000 # higher timeout for debugging

    """SOAP Client"""
    roundtrip_time = observableproperties.ObservableProperty()
    def __init__(self, netloc, logger, sslContext, sdc_definitions, supportedEncodings=None,
                 requestEncodings=None, chunked_requests=False, xml_validator=None):
        """ Connects to one url
        :param netloc: the location of the service (domainname:port) ###url of the service
        :param sslContext: an optional sll.SSLContext instance
        :param supportedEncodings: configured set of encodings that can be used. If None, all available encodings are used.
                                This used for decompression of received responses.
                                If this is an empty list, no compression is supported.
        :param requestEncodings: an optional list of encodings that the other side accepts. It is used to compress requests.
                                If not set, requests will not be commpressed.
                                If set, then the http request will be compressed using this method
        :param xml_validator: optional etree.XMLSchema instance
        """
        self._log = logger
        self._sslContext = sslContext
        self._sdc_definitions = sdc_definitions
        self._netloc = netloc
        self._httpConnection = None # connect later on demand
        self.__class__._usedSoapClients += 1   #pylint: disable=protected-access
        self._clientNo = self.__class__._usedSoapClients   #pylint: disable=protected-access
        self._log.info('created soapClient No. {} for {}', self._clientNo, netloc)
        self.supportedEncodings = supportedEncodings if supportedEncodings is not None else self.available_encodings
        self.requestEncodings = requestEncodings  if requestEncodings is not None else [] # these compression alg's does the other side accept ( set at runtime)
        self._xml_validator = xml_validator
        self._makeGetHeaders()
        self._lock = Lock()
        self._is_closed = True
        self._chunked_requests = chunked_requests
        self.connect()

    @property
    def netloc(self):
        return self._netloc

    @property
    def sock(self):
        return None if self._httpConnection is None else self._httpConnection.sock

    def _mkHttpConnection(self):
        """ Soap client never sends very large requests, the largest packages are notifications.
         Therefore we can use TCP_NODELAY for a little faster transmission.
        (Otherwise there would be a chance that receivers windows size decreases, which would result in smaller
        packages and therefore higher network load."""
        if self._sslContext is not None:
            conn = HTTPSConnection_NODELAY(self._netloc, context=self._sslContext, timeout=self.SOCKET_TIMEOUT)
        else:
            conn =  HTTPConnection_NODELAY(self._netloc, timeout=self.SOCKET_TIMEOUT)
        return conn

    def connect(self):
        with self._lock:
            if self._httpConnection is None:
                self._httpConnection = self._mkHttpConnection()
                self._httpConnection.connect() # connect now so that we have own address and port for logging
                my_addr = self._httpConnection.sock.getsockname()
                self._log.info('soapClient No. {} uses connection={}:{}', self._clientNo, my_addr[0], my_addr[1])
                self._is_closed = False

    def close(self):
        with self._lock:
            if self._httpConnection is not None:
                self._log.info('closing soapClientNo {} for {}', self._clientNo, self._netloc)
                self._httpConnection.close()
                self._httpConnection = None
                self._is_closed = True
    
    def isClosed(self):
        return self._is_closed

    def postSoapEnvelopeTo(self, path,
                           soapEnvelopeRequest,
                           responseFactory=soapenvelope.ReceivedSoap12Envelope.fromXMLString,
                           msg='',
                           request_manipulator=None):
        """
        :param path: url path component
        :param soapEnvelopeRequest: The soap envelope that shall be sent
        :param responseFactory: a callable that creates a response object from received xml. If None, a ReceivedSoap12Envelope will be created
        :param schema: If given, the request is validated against this schema
        :param msg: used in logs, helps to identify the context in which the method was called
        """
        if self.isClosed():
            raise httplib.NotConnected('call connect before posting!')
        return self.__postSoapEnvelope(soapEnvelopeRequest, responseFactory, path, msg, request_manipulator)

    def __postSoapEnvelope(self, soapEnvelopeRequest, responseFactory, path, msg, request_manipulator):
        if self._xml_validator is not None:
            soapEnvelopeRequest.validate_envelope(self._xml_validator)
        if hasattr(request_manipulator, 'manipulate_soapenvelope'):
            tmp = request_manipulator.manipulate_soapenvelope(soapEnvelopeRequest)
            if tmp:
                soapEnvelopeRequest = tmp
        normalized_xml_request = soapEnvelopeRequest.as_xml(request_manipulator=request_manipulator)
        xml_request = self._sdc_definitions.denormalizeXMLText(normalized_xml_request)

        assert (b'utf-8' in xml_request[:100].lower())  # MDPWS:R0007 A text SOAP envelope shall be serialized using utf-8 character encoding
        if hasattr(request_manipulator, 'manipulate_string'):
            tmp = request_manipulator.manipulate_string(xml_request)
            if tmp:
                xml_request = tmp

        started = time.perf_counter()
        try:
            xml_response = self._sendSoapRequest(path, xml_request, msg)
        finally:
            self.roundtrip_time = time.perf_counter() - started # set roundtrip time even if method raises an exception
        normalized_xml_response = self._sdc_definitions.normalizeXMLText(xml_response)

        if responseFactory is not None:
            try:
                response_envelope = responseFactory(normalized_xml_response)
            except etree_.XMLSyntaxError as ex:
                self._log.error('{} XMLSyntaxError in string: "{}"', msg, normalized_xml_response)
                raise RuntimeError('{} in "{}"'.format(ex, normalized_xml_response))
            if self._xml_validator is not None:
                response_envelope.validate_envelope(self._xml_validator)
            return response_envelope
        return None

    def validate_envelope(self, envelope):
        if self._xml_validator is None:
            return
        root = envelope.buildDoc()
        doc = etree_.ElementTree(element=root)
        self._xml_validator.assertValid(doc)

    def _sendSoapRequest(self, path, xml, msg):
        """Send SOAP request using HTTP"""
        if not isinstance(xml, bytes):
            xml = xml.encode('utf-8')

        headers = {
            'Content-type': 'application/soap+xml; charset=utf-8',
            'user_agent': 'pysoap',
            'Connection': 'keep-alive',
        }
        commlog.defaultLogger.logSoapReqOut(xml, 'POST')

        if self.supportedEncodings:
            headers['Accept-Encoding'] = ','.join(self.supportedEncodings)
        if self.requestEncodings:
            for compr in self.requestEncodings:
                if compr in self.supportedEncodings:
                    xml = self.compressPayload(compr, xml)
                    headers['Content-Encoding'] = compr
                    break
        if self._chunked_requests:
            headers['transfer-encoding'] = "chunked"
            xml = mkchunks(xml)
        else:
            headers['Content-Length'] = str(len(xml))

        xml = bytearray(xml)  # cast to bytes, required to bypass httplib checks for is str

        self._log.debug("{}:POST to netloc='{}' path='{}'", msg, self._netloc, path)
        response = None
        content = None

        def send_request():
            do_reopen = False
            success = False
            try:
                self._httpConnection.request('POST', path, body=xml, headers=headers)
                return True, do_reopen # success = True
            except httplib.CannotSendRequest as ex:
                # for whatever reason the response of the previous call was not read. read it and try again
                self._log.warn("{}: could not send request, got httplib.CannotSendRequest Error. Will read response and retry", msg)
                tmp = self._httpConnection.getresponse()
                tmp.read()
            except OSError as ex:
                if ex.errno in (10053, 10054):
                    self._log.warn("{}: could not send request, OSError={!r}", msg, ex)
                else:
                    self._log.warn("{}: could not send request, OSError={}", msg, traceback.format_exc())
                do_reopen = True
            except socket.error as ex:
                self._log.warn("{}: could not send request, socket error={!r}", msg, ex)
                do_reopen = True
            except Exception as ex:
                self._log.warn("{}: POST to netloc='{}' path='{}': could not send request, error={!r}\n{}", msg,
                               self._netloc, path, ex, traceback.format_exc())
            return success, do_reopen

        def get_response():
            try:
                return self._httpConnection.getresponse()
            except OSError as ex:
                if ex.errno in (10053, 10054):
                    self._log.warn("{}: could not receive response, OSError={!r}", msg, ex)
                else:
                    self._log.warn("{}: could not receive response, OSError={} ({!r})\n{}", msg, ex.errno,
                                   ex, traceback.format_exc())
                raise httplib.NotConnected()
            except socket.error as ex:
                self._log.warn("{}: could not receive response, socket error={!r}", msg, ex)
                raise httplib.NotConnected()
            except httplib.BadStatusLine as ex:
                self._log.warn("{}: invalid http response, error= {!r} ", msg, ex)
                raise
            except Exception as ex:
                self._log.warn("{}: POST to netloc='{}' path='{}': could not receive response, error={!r}\n{}",
                               msg, self._netloc, path, ex, traceback.format_exc())
                raise httplib.NotConnected()

        def reopen_http_connection():
            self._log.info("{}: will close and reopen the connection and then try again", msg)
            self._httpConnection.close()
            try:
                self._httpConnection.connect()
            except Exception as ex:
                self._log.error("{}: could not reopen the connection, error={!r}\n{}\ncall-stack ={}",
                                msg, ex, traceback.format_exc(), ''.join(traceback.format_stack()))
                self._httpConnection.close()
                raise httplib.NotConnected()

        with self._lock:
            _retry_send = 2  # ugly construct that allows to retry sending the request once
            while _retry_send > 0:
                _retry_send -= 1
                success, _do_reopen = send_request()
                if not success:
                    if _do_reopen:
                        reopen_http_connection()
                    else:
                        raise httplib.NotConnected()
                else:
                    try:
                        response = get_response()
                        _retry_send = -1  # -1 == SUCCESS
                    except httplib.NotConnected:
                        self._log.info("{}: will reopen after get_response error", msg)
                        reopen_http_connection()

            if _retry_send != -1:
                raise httplib.NotConnected()

            content = HTTPReader.read_response_body(response)

            if response.status >= 300:
                self._log.error(
                    "{}: POST to netloc='{}' path='{}': could not send request, HTTP response={}\ncontent='{}'", msg,
                    self._netloc, path, response.status, content)
                soapfault = soapenvelope.ReceivedSoapFault.fromXMLString(content)

                raise HTTPReturnCodeError(response.status, content, soapfault)

            responseHeaders = {k.lower(): v for k, v in response.getheaders()}

            self._log.debug('{}: response:{}; content has {} Bytes ', msg, responseHeaders, len(content))
            commlog.defaultLogger.logSoapRespIn(content, 'POST')
            return content

    def _makeGetHeaders(self):
        self._getHeaders = {
            'user_agent': 'pysoap',
            'Connection': 'keep-alive'
        }
        if sys.version < '3':
            # Ensure http_method, location and all headers are binary to prevent
            # UnicodeError inside httplib.HTTPConnection._send_output.

            # httplib in python3 do the same inside itself, don't need to convert it here
            self._getHeaders = dict((str(k), str(v)) for k, v in self._getHeaders.items())

        if self.supportedEncodings:
            self._getHeaders['Accept-Encoding'] = ', '.join(self.supportedEncodings)

    def getUrl(self, url, msg):
        if not url.startswith('/'):
            url = '/' + url
        self._log.debug("{} Get {}/{}", msg, self._netloc, url)
        with self._lock:
            self._httpConnection.request('GET', url, headers=self._getHeaders)
            response = self._httpConnection.getresponse()
            headers = {k.lower(): v for k, v in response.getheaders()}
            _content = response.read()
            if 'content-encoding' in headers:
                enc = headers['content-encoding']
                if enc in self.supportedEncodings:
                    content = self.decompress(_content, enc)
                else:
                    self._log.warn("{}: unsupported compression ", headers['content-encoding'])
                    raise httplib.UnknownTransferEncoding
            else:
                content = _content
        return content



from . import commlog
from . import compression
from . import location
from . import mdib
from . import namespaces
from . import netconn
from . import pmtypes
from . import pysoap
from . import safety
from . import sdcclient
from . import sdcdevice
from . import wsdiscovery
from . import xmlparsing

__version__ = '1.3.0'



import logging
import traceback


def ensureLogStream():
    """Method makes sure that the sdc logger has a handler.
    :return: sdc logger
    """
    applog = logging.getLogger('sdc')
    if applog.hasHandlers():
        return
    ch = logging.StreamHandler()
    # create formatter
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    applog.addHandler(ch)
    return applog


def reset_log_levels(root_logger_name='sdc'):
    for name in logging.Logger.manager.loggerDict:
        if name.startswith(root_logger_name):
            logging.getLogger(name).setLevel(logging.NOTSET)


def reset_handlers(root_logger_name='sdc'):
    for name in logging.Logger.manager.loggerDict:
        if name.startswith(root_logger_name):
            logger = logging.getLogger(name)
            for handler in logger.handlers:
                logger.removeHandler(handler)


def basic_logging_setup(root_logger_name='sdc', level=logging.INFO, log_file_name=None):
    logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=level, force=True)
    reset_log_levels(root_logger_name)
    reset_handlers(root_logger_name)
    if log_file_name:
        file_handler = logging.handlers.RotatingFileHandler(log_file_name,
                                                            maxBytes=5000000,
                                                            backupCount=2)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        file_handler.setFormatter(formatter)



class LoggerAdapter(object):
    """
    This adapter wraps a standard logger and changes the interface in two ways:
     - it uses .format() method of strings for formatting (in contrast to logging.Logger, which uses % operator).
     - if any argument in *args or **kwargs is callable, it replaces the argument with the returned value of the called argument.
       This helps to reduce processing time if the called method is expensive and logger is not enabled for given log level.
    """

    def __init__(self, logger, prefix=None):
        self.logger = logger
        self.log_prefix = prefix or ''

    def _process(self, msg, args, kwargs):
        try:
            _msg = self.log_prefix + msg
        except TypeError:
            _msg = msg

        if len(args) == len(kwargs) == 0:
            return _msg

        if '%' in msg and not '{' in msg:
            # traditional log formatting
            return _msg%args

        resolvedArgs = []
        for a in args:
            resolvedArgs.append(a() if callable(a) else a)
        resolvedKwargs = {}
        for k, a in kwargs.items():
            resolvedKwargs[k] = a() if callable(a) else a
        
        try:   
            fullmsg = _msg.format(*resolvedArgs, **resolvedKwargs)
        except:
            print (traceback.format_exc())
            raise
        return fullmsg


    def debug(self, msg, *args, **kwargs):
        self.log(logging.DEBUG, msg, *args, **kwargs)


    def info(self, msg, *args, **kwargs):
        self.log(logging.INFO, msg, *args, **kwargs)


    def warning(self, msg, *args, **kwargs):
        self.log(logging.WARN, msg, *args, **kwargs)

    warn = warning
    
    
    def error(self, msg, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)


    def exception(self, msg, *args, **kwargs):
        self.logger.error(self._process(msg, args, kwargs), exc_info=1)


    def critical(self, msg, *args, **kwargs):
        self.log(logging.CRITICAL, msg, *args, **kwargs)


    def log(self, level, msg, *args, **kwargs):
        """
        Delegate a log call to the underlying logger, after processing msg, args and kwargs
        """
        if self.isEnabledFor(level):
            self.logger.log(level, self._process(msg, args, kwargs))


    def isEnabledFor(self, level):
        """
        See if the underlying logger is enabled for the specified level.
        """
        return self.logger.isEnabledFor(level)


def getLoggerAdapter(name, prefix=None):
    """ Use this method instead of logging.getLogger.
    @return: a LoggerAdapter instance
    """
    return LoggerAdapter(logging.getLogger(name), prefix)


class LogWatchException(Exception):
    def __init__(self, issues):
        super(LogWatchException, self).__init__()
        self.issues = issues

    def __repr__(self):
        return 'LogWatchException: {}'.format(self.issues)


class _LogIssue(object):
    def __init__(self, record):
        self.record = record
        self.call_stack = traceback.format_stack(limit=15)
        # remove last lines from call stack that are inside logging and loghelper.
        # By doing this the call stack shows the call to the logger as last entry.
        while __file__ in self.call_stack[-1] or logging.__file__ in self.call_stack[-1]:
            del self.call_stack[-1]

    def __repr__(self):

        return 'log msg="{}" level={} thread="{}"; call-stack:\n{}'.format(self.record.msg,
                                                                    self.record.levelname,
                                                                    self.record.threadName or self.record.thread,
                                                                    ''.join(self.call_stack))

class LogWatcherHandler(logging.Handler):
    """ This is a logging handler that stores all records in a list"""
    def __init__(self, logger, level):
        """
        This is a logging handler that stores all records in a list.
        :param logger: the logger that shall be handled
        :param level: all records with log level >= level will be recorded
        """
        super(LogWatcherHandler, self).__init__(level=level)
        self._logger = logger
        self.records = []
        self._logger.addHandler(self)

    def emit(self, record):
        """
        This method is called by logger if record log level >= own level
        """
        self.acquire()
        try:
            self.records.append(_LogIssue(record))
        finally:
            self.release()

    def disconnect(self):
        """Remove self from logger."""
        self._logger.removeHandler(self)

    def clear(self):
        """ Delete all records"""
        self.acquire()
        try:
            del self.records[:]
        finally:
            self.release()


class LogWatcher(object):
    """Manages one or more LogWatcherHandlers.
    Can be used also as contextmanager"""
    def __init__(self, logger, level=logging.ERROR, startPaused=False):
        """
        :param logger: the initial logger that shall be recorded
        :param level:  the log level for the initial handler
        :param startPaused: if true, logging is not started immediately.
        """
        self._logger = logger
        self._level = level
        self.handlers = []
        self._collecting = False
        self.addHandler(logger, level)
        self._collecting = not startPaused

    def addHandler(self, logger, level):
        """
        Add another LogWatcherHandler.
        :param logger: the logger that shall be recorded
        :param level: the log level for the handler
        :return: a LogWatcherHandler instance
        """
        coll = LogWatcherHandler(logger, level)
        coll.addFilter(self)
        self.handlers.append(coll)
        return coll

    def setPaused(self, isPaused):
        """
        Enable/disable recording.
        :param isPaused: if True, no records will be saved.
        :return:
        """
        self._collecting = not isPaused

    def stop(self):
        """
        Disconnect and delete all Handlers
        :return:
        """
        self._collecting = False
        for handler in self.handlers:
            handler.disconnect()
        self.handlers = []

    def clearHandlers(self):
        """
        Delete all recorded records in all handlers.
        :return:
        """
        for handler in self.handlers:
            handler.clear()

    def getAllRecords(self):
        """
        :return: a list of all records in all handlers
        """
        all_records = []
        for handler in self.handlers:
            handler.acquire()
            try:
                all_records.extend(handler.records)
            finally:
                handler.release()
        return all_records

    def check(self, stop=True):
        """
        Check for Records. Raises a LogWatchException if any record was found
        :param stop: if True, stop is called internally
        """
        all_records = self.getAllRecords()
        if stop:
            self.stop()
        if all_records:
            raise LogWatchException(all_records)

    def filter(self, record): #pylint: disable=unused-argument
        return self._collecting == True #pylint: disable=singleton-comparision

    def __enter__(self):
        return self

    def __exit__(self, et, ev, tb):
        self.check()



from .. import namespaces
from .. import sdcdevice
from .contextprovider import GenericContextProvider

class GenericPatientContextProvider(GenericContextProvider):

    def __init__(self, log_prefix):
        super(GenericPatientContextProvider, self).__init__(log_prefix)
        self._patientContextDescriptorContainer = None
        self._setPatientContextOperations = []

    def initOperations(self, mdib):
        super(GenericPatientContextProvider, self).initOperations(mdib)
        # expecting exactly one PatientContextDescriptor
        patientContextDescriptorContainers = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('PatientContextDescriptor'))
        if patientContextDescriptorContainers is not None and len(patientContextDescriptorContainers) == 1:
            self._patientContextDescriptorContainer = patientContextDescriptorContainers[0]

    def makeOperationInstance(self, operationDescriptorContainer):
        if self._patientContextDescriptorContainer and operationDescriptorContainer.OperationTarget == self._patientContextDescriptorContainer.handle:
            pc_operation = self._mkOperationFromOperationDescriptor(operationDescriptorContainer,
                                                                    currentArgumentHandler=self._setContextState)
            self._setPatientContextOperations.append(pc_operation)
            return pc_operation

    def makeMissingOperations(self):
        ops = []
        if self._patientContextDescriptorContainer and not self._setPatientContextOperations:
            pc_operation = self._mkOperation(sdcdevice.sco.SetContextStateOperation,
                                             handle='opSetPatCtx',
                                             operationTargetHandle=self._patientContextDescriptorContainer.handle,
                                             codedValue=None,
                                             currentArgumentHandler=self._setContextState)
            ops.append(pc_operation)
        return ops



from sdc11073 import namespaces
from sdc11073 import nomenclature
from sdc11073 import pmtypes
from sdc11073.roles import providerbase
from sdc11073.sdcdevice import sco


class OperationProvider(providerbase.ProviderRole):
    """Handles operations that work on operation states.
     Empty implementation, not needed/used for pysdc tests"""
    pass



import uuid
import copy
from io import BytesIO
from lxml import etree as etree_

from sdc11073.namespaces import wsaTag, wseTag, dpwsTag, s12Tag, xmlTag, nsmap, WSA_ANONYMOUS, docNameFromQName
from sdc11073.namespaces import Prefix_Namespace as Prefix
from .. import isoduration, xmlparsing

CHECK_NAMESPACES = False   # can be used to enable additional checks for too many namespaces or undefined namespaces


DIALECT_ACTION = '{}/Action'.format(Prefix.DPWS.namespace)
DIALECT_THIS_MODEL = '{}/ThisModel'.format(Prefix.DPWS.namespace)
DIALECT_THIS_DEVICE = '{}/ThisDevice'.format(Prefix.DPWS.namespace)
DIALECT_RELATIONSHIP = '{}/Relationship'.format(Prefix.DPWS.namespace)
HOST_TYPE = '{}/host'.format(Prefix.DPWS.namespace)


class SoapResponseException(Exception):
    
    def __init__(self, soapResponseEnvelope):
        super(SoapResponseException, self).__init__()
        self.soapResponseEnvelope = soapResponseEnvelope


class ExtendedDocumentInvalid(etree_.DocumentInvalid):

    pass


def mergeDicts(*args):
    result = {}
    for d in args:
        for k, v in d.items():
            if not k in result:
                result[k] = v 
            else:
                if result[k] != v:
                    raise RuntimeError('Merge Conflict key={}, value1={}, value2={}'.format(k, result[k], v))
    return result

    
def getText(node, idstring, ns):
    if node is None:
        return
    tmp = node.find(idstring, ns)
    if tmp is not None:
        return tmp.text


class GenericNode(object):
    def __init__(self, node):
        self._node = node

        
    def asEtreeSubNode(self, rootNode):
        rootNode.append(self._node)


    
class WsaEndpointReferenceType(object):
    """ Acc. to "http://www.w3.org/2005/08/addressing"

    """
    __slots__ = ('address', 'referenceParametersNode', 'metaDataNode')
    def __init__(self, address, referenceParametersNode=None, metaDataNode=None):
        self.address = address # type="wsa:AttributedURI", which is an xs:anyURI element
        self.referenceParametersNode = None
        self.metaDataNode = None
        if referenceParametersNode is not None:
            if hasattr(referenceParametersNode, 'tag') and referenceParametersNode.tag == wsaTag('ReferenceParameters'):
                self.referenceParametersNode = referenceParametersNode # any content allowed. optional
            else:
                self.referenceParametersNode = etree_.Element(wsaTag('ReferenceParameters'))
                self.referenceParametersNode.extend(referenceParametersNode)
        if metaDataNode is not None:
            if hasattr(metaDataNode, 'tag') and metaDataNode.tag == wsaTag('MetaData'):
                self.metaDataNode = metaDataNode # any content allowed. optional
            else:
                self.metaDataNode = etree_.Element(wsaTag('MetaData'))
                self.metaDataNode.extend(metaDataNode)

    def __str__(self):
        return 'WsaEndpointReferenceType: address={}'.format(self.address)
    
    @classmethod
    def fromEtreeNode(cls, rootNode):
        addressNode = rootNode.find('wsa:Address', nsmap)
        address = addressNode.text
        referenceParametersNode = rootNode.find('wsa:ReferenceParameters', nsmap)
        metaDataNode = rootNode.find('wsa:MetaData', nsmap)
        ret = cls(address, referenceParametersNode, metaDataNode)
        return ret
    
    
    def asEtreeSubNode(self, rootNode):
        node  = etree_.SubElement(rootNode, wsaTag('Address'))
        node.text = self.address
        if self.referenceParametersNode is not None:
            rootNode.append(xmlparsing.copy_node(self.referenceParametersNode))
        if self.metaDataNode is not None:
            rootNode.append(self.metaDataNode)



class WsAddress(object):
    __slots__ = ('messageId', 'to', 'from_', 'replyTo', 'faultTo', 'action',
                 'messageId', 'relatesTo', 'referenceParametersNode', 'relationshipType')
    def __init__(self, action, messageId=None, to=None, relatesTo=None, from_=None, replyTo=None,
                 faultTo=None, referenceParametersNode=None, relationshipType=None): #pylint: disable=too-many-arguments
        """

        :param action: xs:anyURI string, required
        :param messageId: xs:anyURI string or None or False; default is None
                          if None, a messageId is generated automatically
                          if False, no message ID is generated ( makes only sense for testing )
        :param to: xs:anyURI string, optional
        :param relatesTo: xs:anyURI string, 0...n
        :param from_: WsaEndpointReferenceType instance, optional
        :param replyTo: WsaEndpointReferenceType instance, optional
        :param faultTo: WsaEndpointReferenceType instance, optional
        :param referenceParametersNode: any node, optional
        :param relationshipType: a QName, optional
        """
        self.action = action
        if messageId == False:
            self.messageId = None
        else:
            self.messageId = messageId or uuid.uuid4().urn
        self.to = to
        self.relatesTo = relatesTo
        self.from_ = from_
        self.replyTo = replyTo
        self.faultTo = faultTo
        self.referenceParametersNode = referenceParametersNode
        self.relationshipType = relationshipType

    def mkReplyAddress(self, action):
        return WsAddress(action=action, relatesTo=self.messageId)


    def asEtreeSubNode(self, rootNode):
        # To (OPTIONAL), defaults to anonymous
        node = etree_.SubElement(rootNode, wsaTag('To'), attrib={s12Tag('mustUnderstand'): 'true'})
        node.text = self.to or WSA_ANONYMOUS
        #From
        if self.from_:
            self.from_.asEtreeSubNode(rootNode)
        # ReplyTo (OPTIONAL), defaults to anonymous
        if self.replyTo:
            self.replyTo.asEtreeSubNode(rootNode)
        # FaultTo (OPTIONAL)
        if self.faultTo:
            self.faultTo.asEtreeSubNode(rootNode)
        # Action (REQUIRED)
        node  = etree_.SubElement(rootNode, wsaTag('Action'), attrib={s12Tag('mustUnderstand'): 'true'})
        node.text = self.action
        # MessageID (OPTIONAL)
        if self.messageId:
            node  = etree_.SubElement(rootNode, wsaTag('MessageID'))
            node.text = self.messageId
        # RelatesTo (OPTIONAL)
        if self.relatesTo:
            node  = etree_.SubElement(rootNode, wsaTag('RelatesTo'))
            node.text = self.relatesTo
            if self.relationshipType is not None:
                node.set('RelationshipType', self.relationshipType)

        if self.referenceParametersNode:
            rootNode.append(xmlparsing.copy_node(self.referenceParametersNode))

    
    @classmethod
    def fromEtreeNode(cls, rootNode):
        messageId = getText(rootNode, 'wsa:MessageID', nsmap)
        to = getText(rootNode, 'wsa:To', nsmap)
        action = getText(rootNode, 'wsa:Action', nsmap)
        relatesTo = getText(rootNode, 'wsa:RelatesTo', nsmap)
        relationshipType = None
        relatesToNode = rootNode.find('wsa:RelatesTo', nsmap)
        if relatesToNode is not None:
            relatesTo = relatesToNode.text
            relationshipTypeText = relatesToNode.attrib.get('RelationshipType')
            if relationshipTypeText:
                # split into namespace, localname
                ns, loc = relationshipTypeText.rsplit('/', 1)
                relationshipType= etree_.QName(ns, loc)

        def mkEndpointReference(idstring):
            tmp = rootNode.find(idstring, nsmap)
            if tmp is not None:
                return WsaEndpointReferenceType.fromEtreeNode(tmp)
            
        from_ = mkEndpointReference('wsa:From')
        replyTo = mkEndpointReference('wsa:ReplyTo')
        faultTo = mkEndpointReference('wsa:FaultTo')
        referenceParametersNode = rootNode.find('wsa:ReferenceParameters', nsmap)
        
        return cls(messageId=messageId, 
                   to=to, 
                   action=action, 
                   relatesTo=relatesTo, 
                   from_=from_,
                   replyTo=replyTo,
                   faultTo=faultTo,
                   referenceParametersNode=referenceParametersNode,
                   relationshipType=relationshipType)


_LANGUAGE_ATTR = '{http://www.w3.org/XML/1998/namespace}lang'
MODE_PUSH = '{}/DeliveryModes/Push'.format(Prefix.WSE.namespace)


class WsSubscribe(object):
    __slots__ = ('delivery_mode', 'notifyTo',  'endTo', 'expires', 'filter')
    def __init__(self, notifyTo,
                       expires,
                       endTo=None,
                       filter_=None,
                       delivery_mode=None):
        """
        :param notifyTo: a WsaEndpointReferenceType
        :param expires: duration in seconds ( absolute date not supported)
        :param endTo: a WsaEndpointReferenceType or None
        :param delivery_mode: defaults to MODE_PUSH
        """
        self.delivery_mode = delivery_mode or MODE_PUSH
        self.notifyTo = notifyTo
        self.endTo = endTo
        self.expires = expires
        self.filter = filter_


    def asEtreeSubNode(self, rootNode):
        # To (OPTIONAL), defaults to anonymous
        subscribe = etree_.SubElement(rootNode, wseTag('Subscribe'), nsmap=Prefix.partialMap(Prefix.WSE, Prefix.WSA))
        if self.endTo is not None:
            endToNode = etree_.SubElement(subscribe, wseTag('EndTo'))
            self.endTo.asEtreeSubNode(endToNode)
        delivery = etree_.SubElement(subscribe, wseTag('Delivery'))
        delivery.set('Mode', self.delivery_mode)

        notifyToNode  = etree_.SubElement(delivery, wseTag('NotifyTo'))
        self.notifyTo.asEtreeSubNode(notifyToNode)

        exp = etree_.SubElement(subscribe, wseTag('Expires'))
        exp.text = isoduration.durationString(self.expires)
        fil = etree_.SubElement(subscribe, wseTag('Filter'))
        fil.set('Dialect', DIALECT_ACTION) # Is this always this string?
        fil.text= self.filter


    @classmethod
    def fromEtreeNode(cls, rootNode):
        raise NotImplementedError #pylint: disable=unused-argument



class DPWSThisDevice(object):
    __slots__ = ('friendlyName', 'firmwareVersion', 'serialNumber')

    def __init__(self, friendlyName, firmwareVersion, serialNumber):
        if isinstance(friendlyName, dict):
            self.friendlyName = friendlyName
        else:   
            self.friendlyName = {None: friendlyName} # localized texts
        self.firmwareVersion = firmwareVersion
        self.serialNumber = serialNumber


    @classmethod
    def fromEtreeNode(cls, rootNode):
        friendlyName = {} # localized texts
        for m in rootNode.findall('dpws:FriendlyName', nsmap):
            friendlyName[m.get(_LANGUAGE_ATTR)] = m.text 
        firmwareVersion = getText(rootNode, 'dpws:FirmwareVersion', nsmap)
        serialNumber = getText(rootNode, 'dpws:SerialNumber', nsmap)
        return cls(friendlyName, firmwareVersion, serialNumber)


    def asEtreeSubNode(self, rootNode):
        thisDevice = etree_.SubElement(rootNode, dpwsTag('ThisDevice'), nsmap=Prefix.partialMap(Prefix.DPWS))
        for lang, name in self.friendlyName.items():
            friendlyName = etree_.SubElement(thisDevice, dpwsTag('FriendlyName'))
            friendlyName.text = name
            if lang is not None and len(lang) > 0:
                friendlyName.set(_LANGUAGE_ATTR, lang)
        firmwareVersion = etree_.SubElement(thisDevice, dpwsTag('FirmwareVersion'))
        firmwareVersion.text = self.firmwareVersion
        serialNumber = etree_.SubElement(thisDevice, dpwsTag('SerialNumber'))
        serialNumber.text = self.serialNumber


    def __str__(self):
        return 'DPWSThisDevice: friendlyName={}, firmwareVersion="{}", serialNumber="{}"'.format(self.friendlyName, self.firmwareVersion, self.serialNumber)



class DPWSThisModel(object):
    __slots__ = ('manufacturer', 'manufacturerUrl', 'modelName', 'modelNumber', 'modelUrl', 'presentationUrl')
    def __init__(self, manufacturer, manufacturerUrl, modelName, modelNumber, modelUrl, presentationUrl):
        if isinstance(manufacturer, dict):
            self.manufacturer = manufacturer
        else:   
            self.manufacturer = {None: manufacturer} # localized texts
        self.manufacturerUrl = manufacturerUrl
        if isinstance(modelName, dict):
            self.modelName = modelName
        else:   
            self.modelName = {None: modelName} # localized texts
        self.modelNumber = modelNumber
        self.modelUrl = modelUrl
        self.presentationUrl = presentationUrl


    def __str__(self):
        return 'DPWSThisModel: manufacturer={}, modelName="{}", modelNumber="{}"'.format(self.manufacturer, self.modelName, self.modelNumber)


    @classmethod
    def fromEtreeNode(cls, rootNode):
        manufacturer = {} # localized texts
        for m in rootNode.findall('dpws:Manufacturer', nsmap):
            manufacturer[m.get(_LANGUAGE_ATTR)] = m.text 
        manufacturerUrl = getText(rootNode, 'dpws:ManufacturerUrl', nsmap)
        modelName = {} # localized texts
        for m in rootNode.findall('dpws:ModelName', nsmap):
            modelName[m.get(_LANGUAGE_ATTR)] = m.text 
        modelNumber = getText(rootNode, 'dpws:ModelNumber', nsmap)
        modelUrl = getText(rootNode, 'dpws:ModelUrl',  nsmap)
        presentationUrl = getText(rootNode, 'dpws:PresentationUrl', nsmap)
        return cls(manufacturer, manufacturerUrl, modelName, modelNumber, modelUrl, presentationUrl)


    def asEtreeSubNode(self, rootNode):
        thisModel = etree_.SubElement(rootNode, dpwsTag('ThisModel'), nsmap=Prefix.partialMap(Prefix.DPWS))
        for lang, name in self.manufacturer.items():
            manufacturer = etree_.SubElement(thisModel, dpwsTag('Manufacturer'))
            manufacturer.text = name
            if lang is not None and len(lang) > 0:
                manufacturer.set(_LANGUAGE_ATTR, lang)

        manufacturerUrl = etree_.SubElement(thisModel, dpwsTag('ManufacturerUrl'))
        manufacturerUrl.text = self.manufacturerUrl

        for lang, name in self.modelName.items():
            manufacturer = etree_.SubElement(thisModel, dpwsTag('ModelName'))
            manufacturer.text = name
            if lang is not None:
                manufacturer.set(_LANGUAGE_ATTR, lang)

        modelNumber = etree_.SubElement(thisModel, dpwsTag('ModelNumber'))
        modelNumber.text = self.modelNumber
        modelUrl = etree_.SubElement(thisModel, dpwsTag('ModelUrl'))
        modelUrl.text = self.modelUrl
        presentationUrl = etree_.SubElement(thisModel, dpwsTag('PresentationUrl'))
        presentationUrl.text = self.presentationUrl



class DPWSHost(object):
    __slots__ = ('endpointReferences', 'types')
    def __init__(self, endpointReferencesList, typesList):
        """
        :param endpointReferencesList: list of WsEndpointReference instances
        :param typesList: a list of etree.QName instances
        """
        self.endpointReferences = endpointReferencesList
        self.types = typesList


    def asEtreeSubNode(self, rootNode):
        _ns = Prefix.partialMap(Prefix.DPWS, Prefix.WSA)
        # reverse lookup( key is namespace, value is prefix)
        res = {}
        for k,v in _ns.items():
            res[v] = k
        for k,v in rootNode.nsmap.items():
            res[v] = k
        
        # must explicitely add namespaces of types to Host node, because list of qnames is not handled by lxml
        typesTexts = []
        if self.types:
            for qname in self.types:
                prefix = res.get(qname.namespace)
                if not prefix:
                    # create a random prefix
                    prefix='_dpwsh{}'.format(len(_ns))
                    _ns[prefix] = qname.namespace
                typesTexts.append('{}:{}'.format(prefix, qname.localname))
                
        hostNode = etree_.SubElement(rootNode, dpwsTag('Host'))#, nsmap=_ns)
        epRefNode = etree_.SubElement(hostNode, wsaTag('EndpointReference'))#, nsmap=_ns) 
        for epRef in self.endpointReferences:
            epRef.asEtreeSubNode(epRefNode)
            
        if typesTexts:
            typesNode = etree_.SubElement(hostNode, dpwsTag('Types'), nsmap=_ns)# add also namespace prefixes that were locally generated
            typesText = ' '.join(typesTexts)
            typesNode.text = typesText


    @classmethod
    def fromEtreeNode(cls, rootNode):
        endpointReferences = []
        for tmp in rootNode.findall('wsa:EndpointReference', nsmap):
            endpointReferences.append(WsaEndpointReferenceType.fromEtreeNode(tmp))
        types = getText(rootNode, 'dpws:Types', nsmap)
        if types:
            types = types.split()
        return cls(endpointReferences, types)


    def __str__(self):
        return 'DPWSHost: endpointReference={}, types="{}"'.format(self.endpointReferences, self.types)



class DPWSHosted(object):
    __slots__ = ('endpointReferences', 'types', 'serviceId', 'soapClient')
    def __init__(self, endpointReferencesList, typesList, serviceId):
        self.endpointReferences = endpointReferencesList
        self.types = typesList  # a list of QNames
        self.serviceId = serviceId
        self.soapClient = None


    def asEtreeSubNode(self, rootNode):
        hostedNode = etree_.SubElement(rootNode, dpwsTag('Hosted'))
        epRefNode = etree_.SubElement(hostedNode, wsaTag('EndpointReference'))
        for epRef in self.endpointReferences:
            epRef.asEtreeSubNode(epRefNode)
        if self.types:
            typesText = ' '.join([docNameFromQName(t, rootNode.nsmap) for t in self.types])
            typesNode = etree_.SubElement(hostedNode, dpwsTag('Types'))#, nsmap=ns)
            typesNode.text = typesText
        serviceNode = etree_.SubElement(hostedNode, dpwsTag('ServiceId'))#, nsmap=ns)
        serviceNode.text = self.serviceId


    @classmethod
    def fromEtreeNode(cls, rootNode):
        endpointReferences = []
        for tmp in rootNode.findall('wsa:EndpointReference', nsmap):
            endpointReferences.append(WsaEndpointReferenceType.fromEtreeNode(tmp))
        types = getText(rootNode, 'dpws:Types', nsmap)
        if types:
            types = types.split()
        serviceId = getText(rootNode, 'dpws:ServiceId', nsmap)
        return cls(endpointReferences, types, serviceId)

    def __str__(self):
        return 'DPWSHosted: endpointReference={}, types="{}" serviceId="{}"'.format(self.endpointReferences, self.types, self.serviceId)



class DPWSRelationShip(object):
    def __init__(self, rootNode=None):
        hostNode = rootNode.find('dpws:Host', nsmap)
        self.hosted = {}
        self.host = DPWSHost.fromEtreeNode(hostNode)
        for hostedNode in rootNode.findall('dpws:Hosted', nsmap):
            hosted = DPWSHosted.fromEtreeNode(hostedNode)
            self.hosted[hosted.serviceId] = hosted



class MetaDataSection(object):
    def __init__(self, metadataSections):
        self._metadataSections = metadataSections


    def __getattr__(self, attrname):
        try:
            return self._metadataSections[attrname]
        except KeyError:
            raise AttributeError


    @classmethod
    def fromEtreeNode(cls, rootNode):
        metadata = rootNode.find('wsx:Metadata', nsmap)
        metadataSections = {}
        if metadata is not None:
            for metadataSection in metadata.findall('wsx:MetadataSection', nsmap):
                dialect = metadataSection.attrib['Dialect']
                if dialect[-1] == '/': 
                    dialect = dialect[:-1]
                if dialect == "http://schemas.xmlsoap.org/wsdl":
                    locationNode = metadataSection.find('wsx:Location', nsmap)
                    metadataSections['wsdl_location'] = locationNode.text
                elif dialect == DIALECT_THIS_MODEL:
                    thisModelNode = metadataSection.find('dpws:ThisModel', nsmap)
                    metadataSections['thisModel'] = DPWSThisModel.fromEtreeNode(thisModelNode)
                elif dialect == DIALECT_THIS_DEVICE:
                    thisDeviceNode = metadataSection.find('dpws:ThisDevice', nsmap)
                    metadataSections['thisDevice'] = DPWSThisDevice.fromEtreeNode(thisDeviceNode)
                elif dialect == DIALECT_RELATIONSHIP:
                    relationshipNode = metadataSection.find('dpws:Relationship', nsmap)
                    if relationshipNode.get('Type') == HOST_TYPE:
                        metadataSections['relationShip'] = DPWSRelationShip(relationshipNode)
        return cls(metadataSections)


class Soap12EnvelopeBase(object):
    __slots__ = ('_headerNode', '_bodyNode', '_headerObjects', '_bodyObjects', '_docRoot')
    def __init__(self):
        self._headerNode = None
        self._bodyNode = None
        self._headerObjects = []
        self._bodyObjects = []
        self._docRoot = None

    @property
    def headerNode(self):
        return self._headerNode

    @property
    def bodyNode(self):
        return self._bodyNode

    def _assert_valid_exception_wrapper(self, schema, content):
        try:
            schema.assertValid(content)
        except etree_.DocumentInvalid:
            # reformat and validate again to produce better error output
            tmp_str = etree_.tostring(content, pretty_print=True)
            tmp = etree_.parse(BytesIO(tmp_str))
            tmp_str = tmp_str.decode('utf-8')
            try:
                schema.assertValid(tmp)
            except etree_.DocumentInvalid as err:
                msg = "{}\n{}".format(str(err), tmp_str)
                raise ExtendedDocumentInvalid(msg, error_log=err.error_log)


class Soap12Envelope(Soap12EnvelopeBase):
    __slots__ = ('_nsmap', 'address')
    def __init__(self, nsmap):
        super(Soap12Envelope, self).__init__()
        self._nsmap = nsmap
        self.address = None

    def addHeaderObject(self, obj):
        assert hasattr(obj, 'asEtreeSubNode')
        self._headerObjects.append(obj)
        self._docRoot = None

    def addHeaderString(self, headerString):
        element = etree_.fromstring(headerString)
        self.addHeaderObject(GenericNode(element))
        self._docRoot = None
        
    def addHeaderElement(self, element):
        self.addHeaderObject(GenericNode(element))
        self._docRoot = None
    
    def addBodyObject(self, obj):
        assert hasattr(obj, 'asEtreeSubNode')
        self._bodyObjects.append(obj)
        self._docRoot = None

    def addBodyString(self, bodyString):
        element = etree_.fromstring(bodyString)
        self.addBodyObject(GenericNode(element))
        self._docRoot = None

    def addBodyElement(self, element):
        self.addBodyObject(GenericNode(element))
        self._docRoot = None

    def setAddress(self, wsAddress):
        self.address = wsAddress

    def buildDoc(self):
        if self._docRoot is not None:
            return self._docRoot
        
        root = etree_.Element(s12Tag('Envelope'), nsmap=self._nsmap)

        header = etree_.SubElement(root, s12Tag('Header'))
        if self.address:
            self.address.asEtreeSubNode(header)
        for h in self._headerObjects:
            h.asEtreeSubNode(header)
        body = etree_.SubElement(root, s12Tag('Body'))
        for b in self._bodyObjects:
            b.asEtreeSubNode(body)
        self._headerNode = header
        self._bodyNode = body
        self._docRoot = root
        return root

    def as_xml(self, pretty=False, request_manipulator=None):
        tmp = BytesIO()
        root = self.buildDoc()
        doc = etree_.ElementTree(element=root)
        if hasattr(request_manipulator, 'manipulate_domtree'):
            _doc = request_manipulator.manipulate_domtree(doc)
            if _doc:
                doc = _doc
        doc.write(tmp, encoding='UTF-8', xml_declaration=True, pretty_print=pretty)
        return tmp.getvalue()

    def validate_envelope(self, xml_validator):
        if xml_validator is None:
            return
        root = self.buildDoc()
        doc = etree_.ElementTree(element=root)
        xml_validator.assertValid(doc)
        if CHECK_NAMESPACES:
            self._find_unused_namespaces(root)
            self._find_undefined_namespaces()
        self._assert_valid_exception_wrapper(xml_validator, doc)

    def _find_unused_namespaces(self, root):
        xml_doc = self.as_xml()
        unused = []
        used = []
        for prefix, ns in root.nsmap.items():
            _pr = prefix+':'
            if _pr.encode() not in xml_doc:
                unused.append((prefix, ns))
            else:
                used.append(prefix)
        if unused:
            print (root.nsmap, used, xml_doc[:500]) # do not need to see the wohle message
            raise RuntimeError('unused namespaces:{}, used={}'.format(unused, used))

    def _find_undefined_namespaces(self):
        xml_doc = self.as_xml()
        if b':ns0' in xml_doc:
            raise RuntimeError('undefined namespaces:{}'.format(xml_doc))


class ReceivedSoap12Envelope(Soap12EnvelopeBase):
    __slots__ = ('msgNode', 'rawdata', 'address')
    def __init__(self, doc=None, rawdata=None):
        super(ReceivedSoap12Envelope, self).__init__()
        self._docRoot = doc
        self.rawdata = rawdata
        self._headerNode = None
        self._bodyNode = None
        self.address = None
        if doc is not None:
            self._headerNode = doc.find('s12:Header', nsmap)
            self._bodyNode = doc.find('s12:Body', nsmap)
            self.address = WsAddress.fromEtreeNode(self.headerNode)
            try:
                self.msgNode = self.bodyNode[0]
            except IndexError: # body has no content, this can happen
                self.msgNode = None
        

    def as_xml(self, pretty=False):
        tmp = BytesIO()
        doc = etree_.ElementTree(element=self._docRoot)
        doc.write(tmp, encoding='UTF-8', xml_declaration=True, pretty_print=pretty)
        return tmp.getvalue()

    def validate_envelope(self, xml_validator):
        if xml_validator is None:
            return
        self._assert_valid_exception_wrapper(xml_validator, self._docRoot)

    @classmethod
    def fromXMLString(cls, xmlString):
        parser = etree_.ETCompatXMLParser()
        
        try:    
            doc = etree_.fromstring(xmlString, parser=parser)
        except Exception as ex:
            print ('load error "{}" in "{}"'.format(ex, xmlString))
            raise
        return cls(doc=doc, rawdata=xmlString)



class DPWSEnvelope(ReceivedSoap12Envelope):
    __slots__ = ('address', 'thisModel', 'thisDevice', 'hosted', 'host', 'metaData')

    def __init__(self, doc, rawdata):
        super(DPWSEnvelope, self).__init__(doc, rawdata)
        self.address = None
        self.thisModel = None
        self.thisDevice = None
        self.hosted = {}
        self.host = None
        self.metaData = None
        
        if doc is not None:
            self.address = WsAddress.fromEtreeNode(self.headerNode)
            self.metaData = MetaDataSection(self.bodyNode)
            metadata = self.bodyNode.find('wsx:Metadata', nsmap)
            if metadata is not None:
                for metadataSection in metadata.findall('wsx:MetadataSection', nsmap):
                    if metadataSection.attrib['Dialect'] == DIALECT_THIS_MODEL:
                        thisModelNode = metadataSection.find('dpws:ThisModel', nsmap)
                        self.thisModel = DPWSThisModel.fromEtreeNode(thisModelNode)
                    elif metadataSection.attrib['Dialect'] == DIALECT_THIS_DEVICE:
                        thisDeviceNode = metadataSection.find('dpws:ThisDevice', nsmap)
                        self.thisDevice = DPWSThisDevice.fromEtreeNode(thisDeviceNode)
                    elif metadataSection.attrib['Dialect'] == DIALECT_RELATIONSHIP:
                        relationship = metadataSection.find('dpws:Relationship', nsmap)
                        if relationship.get('Type') == HOST_TYPE:
                            hostNode = relationship.find('dpws:Host', nsmap)
                            self.host = DPWSHost.fromEtreeNode(hostNode)
                            for hostedNode in relationship.findall('dpws:Hosted', nsmap):
                                hosted = DPWSHosted.fromEtreeNode(hostedNode)
                                self.hosted[hosted.serviceId] = hosted


class _SoapFaultBase(Soap12Envelope):
    """
    created xml:
        <S:Body>
            <S:Fault>
                <S:Code>
                    <S:Value>[code]</S:Value>
                    <S:Subcode>
                        <S:Value>[subcode]</S:Value>
                    </S:Subcode>
                </S:Code>
                <S:Reason>
                    <S:Text xml:lang="en">[reason]</S:Text>
                </S:Reason>
                <S:Detail>
                    [detail]
                </S:Detail>
            </S:Fault>
        </S:Body>

    """
    def __init__(self, requestEnvelope, fault_action, code, reason, subCode, details):
        super(_SoapFaultBase, self).__init__(Prefix.partialMap(Prefix.S12, Prefix.WSA,Prefix.WSE))
        replyAddress = requestEnvelope.address.mkReplyAddress(fault_action)
        self.addHeaderObject(replyAddress)
        faultNode = etree_.Element(s12Tag('Fault'))
        codeNode = etree_.SubElement(faultNode, s12Tag('Code'))
        valueNode = etree_.SubElement(codeNode, s12Tag('Value'))
        valueNode.text = 's12:{}'.format(code)
        if subCode is not None:
            subcodeNode = etree_.SubElement(codeNode, s12Tag('Subcode'))
            valueNode = etree_.SubElement(subcodeNode, s12Tag('Value'))
            valueNode.text = docNameFromQName(subCode, nsmap)
        reasonNode = etree_.SubElement(faultNode, s12Tag('Reason'))
        reasontextNode = etree_.SubElement(reasonNode, s12Tag('Text'))
        reasontextNode.set(xmlTag('lang'), 'en-US')
        reasontextNode.text = reason
        if details is not None:
            _detailNode = etree_.SubElement(faultNode, s12Tag('Detail'))
            _detailNode.set(xmlTag('lang'), 'en-US')
            if isinstance(details, str):
                detNode = etree_.SubElement(_detailNode, 'data')
                detNode.text = details
            else:
                _detailNode.append(details)
        self.addBodyElement(faultNode)


class SoapFault(_SoapFaultBase):
    SOAP_FAULT_ACTION = '{}/soap/fault'.format(Prefix.WSA.namespace)
    def __init__(self, requestEnvelope, code, reason, subCode=None, details=None):
        super(SoapFault, self).__init__(requestEnvelope, self.SOAP_FAULT_ACTION, code, reason, subCode, details)


class AdressingFault(_SoapFaultBase):
    ADDRESSING_FAULT_ACTION = '{}/fault'.format(Prefix.WSA.namespace)
    def __init__(self, requestEnvelope, code, reason, subCode=None, details=None):
        super(AdressingFault, self).__init__(requestEnvelope, self.ADDRESSING_FAULT_ACTION, code, reason, subCode, details)


class ReceivedSoapFault(ReceivedSoap12Envelope):
    def __init__(self, doc=None, rawdata=None):
        super(ReceivedSoapFault, self).__init__(doc, rawdata)
        self.code = ', '.join(self._bodyNode.xpath('s12:Fault/s12:Code/s12:Value/text()', namespaces=nsmap))
        self.subcode = ', '.join(self._bodyNode.xpath('s12:Fault/s12:Code/s12:Subcode/s12:Value/text()', namespaces=nsmap))
        self.reason = ', '.join(self._bodyNode.xpath('s12:Fault/s12:Reason/s12:Text/text()', namespaces=nsmap))
        self.detail = ', '.join(self._bodyNode.xpath('s12:Fault/s12:Detail/text()', namespaces=nsmap))

    def __repr__(self):
        return ('ReceivedSoapFault(code="{}", subcode="{}", reason="{}", detail="{}")'.format(self.code, self.subcode, self.reason, self.detail))


class SoapFaultCode:
    """
        Soap Fault codes, see https://www.w3.org/TR/soap12-part1/#faultcodes
    """
    VERSION_MM = 'VersionMismatch'
    MUSTUNSERSTAND = 'MustUnderstand'
    DATAENC = 'DataEncodingUnknown'
    SENDER = 'Sender'
    RECEIVER = 'Receiver'



import copy
import inspect
from lxml import etree as etree_
from .. import observableproperties as properties, xmlparsing
from ..namespaces import QN_TYPE
from ..namespaces import Prefix_Namespace as Prefix

class ContainerBase(object):
    NODETYPE = None   # overwrite in derived classes! determines the value of xsi:Type attribute, must be a etree_.QName object
    NODENAME = None
    node = properties.ObservableProperty()

    # every class with containerproperties must provice a list of property names.
    # this list is needed to create sub elements in a certain order.
    # rule is : elements are sorted from this root class to derived class. Last derived class comes last.
    # Initialization is from left to right
    # This is according to the inheritance in BICEPS xml schema
    _props = tuple()  # empty tuple, this base class has no properties

    def __init__(self, nsmapper, node=None):
        self.nsmapper = nsmapper
        self.node = node
        if node is None:
            # initialize all ContainerProperties
            for dummy_name, cprop in self._sortedContainerProperties():
                cprop.initInstanceData(self)
        else:
            self._updateFromNode(node)


    def getActualValue(self, attr_name):
        """ ignores default value and implied value, e.g. returns None if value is not present in xml"""
        return getattr(self.__class__, attr_name).getActualValue(self)


    def mkNode(self, tag=None, setXsiType=False):
        """
        create a etree node from instance data
        :param tag: tag of the newly created node, defaults to self.NODENAME
        :return: etree node
        """
        myTag = tag or self.NODENAME
        node = etree_.Element(myTag, nsmap=self.nsmapper.partialMap(Prefix.PM, Prefix.MSG, Prefix.XSI))
        self._updateNode(node, setXsiType)
        return node


    def _updateNode(self, node, setXsiType=False):
        """
        create a etree node from instance data
        :param tag: tag of the newly created node, defaults to self.NODENAME
        :return: etree node
        """
        if setXsiType and self.NODETYPE is not None:
            node.set(QN_TYPE, self.nsmapper.docNameFromQName(self.NODETYPE))
        for dummy_name, prop in self._sortedContainerProperties():
            prop.updateXMLValue(self, node)
        return node


    def _updateFromNode(self, node):
        """ update members.
        """
        # update all ContainerProperties
        for dummy_name, cprop in self._sortedContainerProperties():
            cprop.updateFromNode(self, node)

    def mkCopy(self, copy_node=True):
        copied = copy.copy(self)  # no deepcopy because of TypeError: cannot pickle 'lxml.etree.QName' object
        if copy_node and self.node is not None:
            new_node = xmlparsing.copy_node(self.node)
            copied.node = new_node
        return copied

    def _sortedContainerProperties(self):
        """
        @return: a list of (name, object) tuples of all GenericProperties ( and subclasses)
        """
        ret = []
        all_classes = inspect.getmro(self.__class__)
        for cls in reversed(all_classes):
            try:
                names = cls._props  # pylint:disable=protected-access
            except:
                continue
            for name in names:
                obj = getattr(cls, name)
                if obj is not None:
                    ret.append((name, obj))
        return ret


    def diff(self, other, ignore_property_names=None):
        """ compares all properties (except to be ignored ones).
        returns a list of strings that describe differences"""
        ret = []
        ignore_list = ignore_property_names or []
        my_properties = self._sortedContainerProperties()
        for name, dummy in my_properties:
            if name in ignore_list:
                continue
            my_value = getattr(self, name)
            try:
                other_value = getattr(other, name)
            except AttributeError:
                ret.append('{}={}, other does not have this attribute'.format(name, my_value))
            else:
                if isinstance(my_value, float) or isinstance(other_value, float):
                    # cast both to float, if one is a Decimal Exception might be thrown
                    if abs((float(my_value)-float(other_value))/float(my_value)) > 1e-10: # 1e-10 is good enough
                        ret.append('{}={}, other={}'.format(name, my_value, other_value))
                elif my_value != other_value:
                    ret.append('{}={}, other={}'.format(name, my_value, other_value))
        # check also if other has a different list of properties
        my_property_names = set([p[0] for p in my_properties])
        other_property_names = set([p[0] for p in other._sortedContainerProperties()])
        surplus_names = other_property_names - my_property_names
        if surplus_names:
            ret.append(f'other has more data elements:{surplus_names}')
        return ret

    def is_equal(self, other):
        return len(self.diff(other)) == 0

    def __repr__(self):
        return '{} name="{}" type={}'.format(self.__class__.__name__, self.NODENAME, self.NODETYPE)




import re
from datetime import timedelta
from datetime import datetime, date, tzinfo
from decimal import Decimal
from collections import namedtuple

ISO8601_PERIOD_REGEX = re.compile(
    r"^(?P<sign>[+-])?"
    r"P(?!\b)"
    r"(?P<years>[0-9]+([,.][0-9]+)?Y)?"
    r"(?P<months>[0-9]+([,.][0-9]+)?M)?"
    r"(?P<weeks>[0-9]+([,.][0-9]+)?W)?"
    r"(?P<days>[0-9]+([,.][0-9]+)?D)?"
    r"((?P<separator>T)(?P<hours>[0-9]+([,.][0-9]+)?H)?"
    r"(?P<minutes>[0-9]+([,.][0-9]+)?M)?"
    r"(?P<seconds>[0-9]+([,.][0-9]+)?S)?)?$")
# regular expression to parse ISO duration strings.


def parse_duration(datestring):
    """
    Parses an ISO 8601 durations into a float value containing seconds.
    The following duration formats are supported:
      -PnnW                  duration in weeks
      -PnnYnnMnnDTnnHnnMnnS  complete duration specification
    Years and month are not supported, values must be zero! 
    """
    if not isinstance(datestring, str):
        raise TypeError("Expecting a string %r" % datestring)
    match = ISO8601_PERIOD_REGEX.match(datestring)
    if not match:
        raise ValueError("Unable to parse duration string %r" % datestring)
    groups = match.groupdict()
    for key, val in groups.items():
        if key not in ('separator', 'sign'):
            if val is None:
                groups[key] = "0n"
            # print groups[key]
            if key in ('years', 'months'):
                groups[key] = Decimal(groups[key][:-1].replace(',', '.'))
            else:
                # these values are passed into a timedelta object,
                # which works with floats.
                groups[key] = float(groups[key][:-1].replace(',', '.'))
    if groups["years"] != 0 or groups["months"] != 0:
        raise ValueError("Unable to parse duration string %r (Non zero year or month)" % datestring)
    else:
        ret = timedelta(days=groups["days"], hours=groups["hours"],
                        minutes=groups["minutes"], seconds=groups["seconds"],
                        weeks=groups["weeks"])
        if groups["sign"] == '-':
            ret = timedelta(0) - ret
        return ret.total_seconds()


def durationString(seconds):
    sign = '-' if seconds < 0 else ''
    fract = abs(seconds - int(seconds))
    seconds = abs(int(seconds))
    minutes, sec = divmod(seconds,60)
    hours, minutes= divmod(minutes,60)
    days, hours = divmod(hours, 24)
#    sec += fract
    if fract == 0:
        fract_string = ''
    else:
        fract_string = f'{fract:.9f}'[1:] # starting from dot char
        while fract_string[-1] == '0':
            fract_string = fract_string[:-1]
    if days == 0:
        return '{}PT{}H{}M{}{}S'.format(sign, hours, minutes, sec, fract_string)
    else:
        return '{}P0Y0M{}DT{}H{}M{}{}S'.format(sign, days, hours, minutes, sec, fract_string)


##### Date Time ######
GYearMonth = namedtuple('GYearMonth', 'year month')
GYear = namedtuple('GYear', 'year')

ZERO = timedelta(0)
class UTC(tzinfo):
    """Fixed offset in minutes east from UTC."""
    def __init__(self, offset_minutes, tzname=None):
        self._offset = timedelta(minutes=offset_minutes)
        self._tzname = tzname

    def utcoffset(self, dt): #pylint:disable=unused-argument
        return self._offset

    def tzname(self, dt): #pylint:disable=unused-argument
        return self._tzname

    def dst(self, dt): #pylint:disable=unused-argument
        return ZERO


# regular expression to parse ISO 8601 date / datetime strings.

_tz_regex_str = '(?P<tz>Z|((?P<tz_sign>[+-])(?P<tz_hours>[0-9]{1,2}):(?P<tz_minutes>[0-9]{1,2})))?'
_date_regex_str = '(?P<year>[0-9]{4})-?(?P<month>1[0-2]|0[1-9])-?(?P<day>[0-9]{1,2})'
_time_regex_str = '(?P<hour>2[0-3]|[01][0-9]):?(?P<minute>[0-5][0-9]):(?P<second>[0-5][0-9]([.][0-9]+)?)'
_datetime_regex = re.compile('^'+_date_regex_str + '(T' + _time_regex_str + _tz_regex_str + ')?')
_datetime_regex_relaxed = re.compile('^'+_date_regex_str + '([T, ]' + _time_regex_str + _tz_regex_str + ')?') # allows space between date and time
_year_month_regex = re.compile('^(?P<year>[0-9]{4})(-(?P<month>1[0-2]|0[1-9]))?')


def parse_date_time(date_time_string, strict=True):
    try:
        if strict:
            d = _datetime_regex.match(date_time_string)
        else:
            d = _datetime_regex_relaxed.match(date_time_string)
        if d is not None:
            groups = d.groupdict()
            year, month, day = int(groups['year']), int(groups['month']), int(groups['day'])
            if groups['hour'] is None: # only a date, no time
                return date(year, month, day)

            tz_1st = groups['tz']
            if tz_1st is None:
                tz_info = None
            elif tz_1st == 'Z':
                tz_info = UTC(0, 'UTC')
            elif tz_1st[0] in ('+', '-'):
                tz_hours = int(groups['tz_hours'])
                tz_minutes = int(groups['tz_minutes'])
                offset = tz_hours*60+tz_minutes
                if tz_1st[0] == '-':
                    offset *= -1
                tz_info = UTC(offset, 'unknown')

            hour = int(groups['hour'])
            minute = int(groups.get('minute', '00'))
            second = float(groups.get('second', '0.0'))
            sec, microsec = int(second), int((second - int(second))*1000000)
            value = datetime(year, month, day, hour, minute, sec, microsec, tz_info)
            return value

        d = _year_month_regex.match(date_time_string)
        if d is not None:
            groups = d.groupdict()
            year, month = groups['year'], groups['month']
            if month is None:
                value = GYear(int(year))
            else:
                value = GYearMonth(int(year), int(month))
            return value

        raise ValueError('Could not parse date string = "{}"'.format(date_time_string))
    except ValueError:
        return None


def _mkSecondsString(date_object):
    if date_object.microsecond > 0:
        seconds = float(date_object.second) + float(date_object.microsecond)/1e6
        secondsString = '{:06.03f}'.format(seconds)
        #remove trailing zeros
        while secondsString[-1] == '0':
            secondsString = secondsString[:-1]
    else:
        secondsString = '{:02d}'.format(date_object.second)
    return secondsString

def _mkTzString(date_object):
    tz_string = ''
    if date_object.tzinfo:
        delta = date_object.tzinfo.utcoffset(0)
        tz_seconds = delta.seconds + (3600 * 24) * delta.days
        if tz_seconds == 0:
            tz_string = 'Z'
        if tz_seconds != 0:
            minutes, sec = divmod(abs(tz_seconds), 60)
            hours, minutes = divmod(minutes, 60)
            tz_string = '{}{:02d}:{:02d}'.format('+' if tz_seconds > 0 else '-', hours, minutes)
    return tz_string

def date_time_string(date_object):
    if hasattr(date_object, 'hour'): # datetime object
        datestring = '{:4d}-{:02d}-{:02d}T{:02d}:{:02d}:{}{}'.format(date_object.year,
                                                                     date_object.month,
                                                                     date_object.day,
                                                                     date_object.hour,
                                                                     date_object.minute,
                                                                     _mkSecondsString(date_object),
                                                                     _mkTzString(date_object))
    elif hasattr(date_object, 'day'): #date object
        datestring = '{:4d}-{:02d}-{:02d}'.format(date_object.year,
                                                  date_object.month,
                                                  date_object.day)
    elif hasattr(date_object, 'month'): #GYearMonth object
        datestring = '{:4d}-{:02d}'.format(date_object.year, date_object.month)
    elif hasattr(date_object, 'year'): #GYear object
        datestring = '{:4d}'.format(date_object.year)
    else:
        raise ValueError('cannot convert {} to ISO8601 datetime string'.format(date_object.__class__.__name__))
    return datestring


# -*- coding: utf-8 -*-
"""Container properties represent values in xml nodes.
These values can be node attributes, node texts or a complete Elements with optional sub nodes.
The properties completely hide the XML nature of data. To serve this purpose, they can convert between XML data types and Python data types.
"""
import copy
import datetime
import time

from lxml import etree as etree_

import sdc11073.namespaces as namespaces
from sdc11073 import isoduration
from sdc11073.dataconverters import TimestampConverter, DecimalConverter, IntegerConverter, BooleanConverter, \
    DurationConverter, NullConverter
from sdc11073.xmlparsing import copy_node_wo_parent


class ElementNotFoundException(Exception):
    pass


class _PropertyValue:
    """This class contains two representations of a value (XML side and Python side) """

    def __init__(self, xml_value, py_value):
        self.xml_value = xml_value
        self.py_value = py_value


class _PropertyBase(object):
    """ Navigates to sub element and handles storage of value in instance.

    All Properties have the same interface:
    __get__ and __set__ : read and write access, using Python data types.
    getPyValueFromNode: reads the value from XML data and converts it to Python data type.
    updateXMLValue: convert the Python data type to XML type and write it to XML node.
     """

    def __init__(self, attrname, subElementNames, defaultPyValue, impliedPyValue=None):
        """

        :param attrname: name of the attribute that an instance represents
        :param subElementNames: a list of element names that define the path to the attribute in a node (can be an empty list)
        :param defaultPyValue: initial value when initialized (should be set for mandatory elements, otherwise created xml might violate schema)
                               and if the xml element does not exist.
        :param impliedPyValue: for optional elements, this is the value that shall be implied if xml element does not exist
                                this value is for information only! Access only via class possible.
        """
        self._attrname = attrname
        if subElementNames is None:
            self._subElementNames = []
        else:
            self._subElementNames = subElementNames
        if subElementNames is not None:
            localVarName = '_' + '_'.join([s.localname.lower() for s in subElementNames])
        else:
            localVarName = ''
        if attrname is not None:  # add attrname
            if isinstance(attrname, etree_.QName):
                localVarName = localVarName + '_' + attrname.localname
            else:
                localVarName = localVarName + '_' + attrname.lower()
        self._localVarName = localVarName
        self._defaultPyValue = defaultPyValue
        self.impliedPyValue = impliedPyValue

    def __get__(self, instance, owner):
        """ returns a python value, uses the locally stored value"""
        if instance is None:  # if called via class
            return self
        try:
            property_value = getattr(instance, self._localVarName)
            value = property_value.py_value
        except AttributeError:
            value = None
        if value is None:
            if callable(self.impliedPyValue):
                value = self.impliedPyValue()
            else:
                value = self.impliedPyValue
        return value

    def getActualValue(self, instance):
        """ Returns the actual value without considering default value and implied value, e.g. returns None if no value in xml exists."""
        try:
            return getattr(instance, self._localVarName).py_value
        except AttributeError:
            return None

    def __set__(self, instance, pyValue):
        """value is the representation on the program side, e.g a float. """
        if pyValue is None:
            setattr(instance, self._localVarName, None)
        else:
            setattr(instance, self._localVarName, _PropertyValue(None, pyValue))

    def initInstanceData(self, instance):
        if self._defaultPyValue is None:
            setattr(instance, self._localVarName, None)
        else:
            value = copy.copy(self._defaultPyValue)
            setattr(instance, self._localVarName, _PropertyValue(None, value))

    def updateXMLValue(self, instance, node):
        # to be defined in derived classes
        raise NotImplementedError

    def getPyValueFromNode(self, node):
        # to be defined in derived classes
        raise NotImplementedError

    @staticmethod
    def _getElementbyChildNamesList(node, subElementNames, createMissingNodes):
        for n in subElementNames:
            subNode = node.find(n)
            if subNode is None:
                if not createMissingNodes:
                    raise ElementNotFoundException(
                        'Element {} not found in {}, path={}'.format(n, node.tag, subElementNames))
                subNode = etree_.SubElement(node, n)  # create this node
            node = subNode
        return node

    def rmLastSubElement(self, node):
        try:
            pNode = self._getElementbyChildNamesList(node, self._subElementNames[:-1], createMissingNodes=False)
        except ElementNotFoundException:
            return
        subNode = pNode.find(self._subElementNames[-1])
        if subNode is not None:
            pNode.remove(subNode)

    def updateFromNode(self, instance, node):
        value = self.getPyValueFromNode(node)
        setattr(instance, self._localVarName, value)

    def __str__(self):
        if self._subElementNames:
            path_string = ', '.join([str(x) for x in self._subElementNames])
            return '{} attribute {} in subelement {}'.format(self.__class__.__name__, self._attrname, path_string)
        else:
            return '{} attribute {}'.format(self.__class__.__name__, self._attrname)


class NotImplementedProperty(_PropertyBase):
    """ For place holders """

    def __init__(self, attrname, subElementNames=None):
        super(NotImplementedProperty, self).__init__(attrname, subElementNames, None)

    def __get__(self, instance, owner):
        return None

    def __set__(self, instance, value):
        raise NotImplementedError


class _ListPropertyBase(_PropertyBase):
    """ Base class for all classes that have an empty list as default value.
    These classes do not use an implied value.
    The local variable is just a plain list, no _PropertyValue"""

    def __init__(self, attrname, subElementNames):
        super().__init__(attrname, subElementNames, None)

    def __set__(self, instance, pyValue):
        """value is the representation on the program side, e.g a float. """
        setattr(instance, self._localVarName, pyValue)

    def __get__(self, instance, owner):
        """ returns a python value, uses the locally stored value"""
        if instance is None:  # if called via class
            return self
        try:
            return getattr(instance, self._localVarName)
        except AttributeError:
            self.initInstanceData(instance)
            return getattr(instance, self._localVarName)

    def initInstanceData(self, instance):
        setattr(instance, self._localVarName, [])


class NodeAttributeProperty(_PropertyBase):
    """ XML Representation is a string, Python representation is determined by valueConverter."""

    def __init__(self, attrname, subElementNames=None, valueConverter=None, defaultPyValue=None, impliedPyValue=None):
        super(NodeAttributeProperty, self).__init__(attrname, subElementNames, defaultPyValue, impliedPyValue)
        self._converter = valueConverter if valueConverter is not None else NullConverter

    def getPyValueFromNode(self, node):
        value = self._defaultPyValue
        xmlValue = None
        try:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
            xmlValue = subNode.attrib.get(self._attrname)
            if xmlValue is not None:
                value = self._converter.toPy(xmlValue)
        except ElementNotFoundException:
            pass
        if xmlValue is None and value is None:
            return None
        return _PropertyValue(xmlValue, value)

    def updateXMLValue(self, instance, node):
        try:
            property_value = getattr(instance, self._localVarName)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            property_value = None
        if property_value is None:
            try:
                subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
                if subNode is None:
                    return
                if self._attrname in subNode.attrib.keys():
                    del subNode.attrib[self._attrname]
            except ElementNotFoundException:
                return
        else:
            # use xml_value if available, otherwise convert py_value to xml_value
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
            if property_value.xml_value is not None:
                subNode.set(self._attrname, property_value.xml_value)
            elif property_value.py_value is not None:
                xml_value = self._converter.toXML(property_value.py_value)
                if xml_value is not None:
                    subNode.set(self._attrname, xml_value)


class NodeAttributeListProperty(_ListPropertyBase):
    """ XML Representation is a string which is a space separated list"""

    def __init__(self, attrname, subElementNames=None, valueConverter=None):
        super(NodeAttributeListProperty, self).__init__(attrname, subElementNames)
        self._converter = valueConverter if valueConverter is not None else NullConverter

    def getPyValueFromNode(self, node):
        value = self._defaultPyValue
        try:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
            xmlValue = subNode.attrib.get(self._attrname)
            if xmlValue is not None:
                value = [h for h in self._converter.toPy(xmlValue).split(' ') if h]
        except ElementNotFoundException:
            pass
        return value

    def updateXMLValue(self, instance, node):
        try:
            pyValue = getattr(instance, self._localVarName)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            pyValue = None
        if pyValue is None:
            try:
                subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
                if subNode is None:
                    return
                if self._attrname in subNode.attrib.keys():
                    del subNode.attrib[self._attrname]
            except ElementNotFoundException:
                return
        else:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
            xmlValue = ' '.join(self._converter.toXML(pyValue))
            subNode.set(self._attrname, xmlValue)


class XsiTypeAttributeProperty(_PropertyBase):
    """ XML Representation is a namespace:name string, Python representation is a QName."""

    def __init__(self, attrname, subElementNames=None, defaultPyValue=None, impliedPyValue=None):
        super(XsiTypeAttributeProperty, self).__init__(attrname, subElementNames, defaultPyValue, impliedPyValue)

    def getPyValueFromNode(self, node):
        """
        :param node: the etree node as input
        @return: None or a QName
        """
        value = self._defaultPyValue
        try:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
            xmlValue = subNode.attrib.get(self._attrname)
            if xmlValue is not None:
                value = namespaces.txt2QName(xmlValue, node.nsmap)
        except ElementNotFoundException:
            pass
        if xmlValue is None and value is None:
            return None
        return _PropertyValue(xmlValue, value)

    def updateXMLValue(self, instance, node):
        try:
            property_value = getattr(instance, self._localVarName)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            property_value = None
        if property_value is None:
            try:
                subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
                if self._attrname in subNode.attrib.keys():
                    del subNode.attrib[self._attrname]
            except ElementNotFoundException:
                return
        else:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
            if property_value.xml_value is not None:
                xml_value = property_value.xml_value
            elif property_value.py_value is not None:
                xml_value = namespaces.docNameFromQName(property_value.py_value, node.nsmap)
            subNode.set(self._attrname, xml_value)


class DecimalListAttributeProperty(_ListPropertyBase):
    """ XML representation: an attribute string that represents 1..n decimals, separated with spaces.
        Python representation: a list of integers and/or floats.
        """

    def __init__(self, attrname, subElementNames=None):
        super(DecimalListAttributeProperty, self).__init__(attrname, subElementNames)

    def getPyValueFromNode(self, node):
        try:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
            xmlValue = subNode.attrib.get(self._attrname)
            if xmlValue is not None:
                xmlValues = xmlValue.split()
                values = [DecimalConverter.toPy(v) for v in xmlValues]
                return values
        except ElementNotFoundException:
            pass
        return self._defaultPyValue

    def updateXMLValue(self, instance, node):
        try:
            value = getattr(instance, self._localVarName)
        except AttributeError:
            value = None
        # value is a list of integer/float or None
        if value is None:
            try:
                subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
                if self._attrname in subNode.attrib.keys():
                    del subNode.attrib[self._attrname]
            except ElementNotFoundException:
                return
            return
        else:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
            attr_value = ' '.join([DecimalConverter.toXML(v) for v in value])
            subNode.set(self._attrname, attr_value)


class NodeTextProperty(_PropertyBase):
    """ The handled data is the text of an element."""

    def __init__(self, subElementNames=None, valueConverter=None, defaultPyValue=None, impliedPyValue=None,
                 isOptional=True):
        attrname = '_text'
        super(NodeTextProperty, self).__init__(attrname, subElementNames, defaultPyValue, impliedPyValue)
        self._converter = valueConverter if valueConverter is not None else NullConverter
        self._isOptional = isOptional

    def getPyValueFromNode(self, node):
        try:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
            xmlValue = subNode.text
            if xmlValue is not None:
                value = self._converter.toPy(xmlValue)
                return _PropertyValue(xmlValue, value)
        except ElementNotFoundException:
            pass
        if self._defaultPyValue is None:
            return None
        return _PropertyValue(None, self._defaultPyValue)

    def updateXMLValue(self, instance, node):
        try:
            property_value = getattr(instance, self._localVarName)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            property_value = None
        if property_value is None:
            try:
                parentNode = self._getElementbyChildNamesList(node, self._subElementNames[:-1],
                                                              createMissingNodes=False)
            except ElementNotFoundException:
                return
            if not self._subElementNames:
                # update text of this element
                node.text = ''
            else:
                if self._isOptional:
                    subNode = parentNode.find(self._subElementNames[-1])
                    if subNode is not None:
                        parentNode.remove(subNode)
                else:
                    subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
                    subNode.text = None
        else:
            value = None
            if property_value.xml_value is not None:
                value = property_value.xml_value
            elif property_value.py_value is not None:
                value = self._converter.toXML(property_value.py_value)
            if value is not None:
                subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
                subNode.text = value

    def __str__(self):
        if self._subElementNames:
            path_string = ', '.join([str(x) for x in self._subElementNames])
            return '{} in subelement {}'.format(self.__class__.__name__, path_string)
        else:
            return '{} '.format(self.__class__.__name__, )


class NodeTextQNameProperty(_PropertyBase):
    """ The handled data is a qualified name as in the text of an element"""

    def __init__(self, subElementNames, defaultPyValue=None):
        attrname = None
        super(NodeTextQNameProperty, self).__init__(attrname, subElementNames, defaultPyValue)

    def getPyValueFromNode(self, node):
        try:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
            xmlValue = subNode.text
            if xmlValue is not None:
                value = namespaces.txt2QName(xmlValue, subNode.nsmap)
                return _PropertyValue(xmlValue, value)
        except ElementNotFoundException:
            pass
        if self._defaultPyValue is None:
            return None
        return _PropertyValue(None, self._defaultPyValue)

    def updateXMLValue(self, instance, node):
        try:
            property_value = getattr(instance, self._localVarName)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            property_value = None

        if property_value is None:
            try:
                parentNode = self._getElementbyChildNamesList(node, self._subElementNames[:-1],
                                                              createMissingNodes=False)
            except ElementNotFoundException:
                return
            subNode = parentNode.find(self._subElementNames[-1])
            if subNode is not None:
                parentNode.remove(subNode)
        else:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
            if property_value.xml_value is not None:
                value = property_value.xml_value
            else:
                value = namespaces.docNameFromQName(property_value.py_value, subNode.nsmap)
            subNode.text = value


def _compare_extension(left: etree_.ElementBase, right: etree_.ElementBase) -> bool:
    # xml comparison
    try:
        if left.tag != right.tag:  # compare expanded names
            return False
        if dict(left.attrib) != dict(right.attrib):  # unclear how lxml _Attrib compares
            return False
    except AttributeError:  # right side is not an Element type because expected attributes are missing
        return False

    # ignore comments
    left_children = [child for child in left if not isinstance(child, etree_._Comment)]
    right_children = [child for child in right if not isinstance(child, etree_._Comment)]

    if len(left_children) != len(right_children):  # compare children count
        return False
    if len(left_children) == 0 and len(right_children) == 0:
        if left.text != right.text:  # mixed content is not allowed. only compare text if there are no children
            return False

    return all(map(_compare_extension, left_children, right_children))  # compare children but keep order


class ExtensionLocalValue(list):

    compare_method = _compare_extension
    """may be overwritten by user if a custom comparison behaviour is required"""

    def __eq__(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:  # len of other cannot be determined
            return False
        return all(self.__class__.compare_method(left, right) for left, right in zip(self, other))


class ExtensionNodeProperty(_PropertyBase):
    """Represents an ext:Extension Element that contains 0...n child elements of any kind.

    The python representation is an ExtensionLocalValue with list of elements.
    """

    def __init__(self, subElementNames=None, defaultPyValue=None):
        if subElementNames is None:
            subElementNames = [namespaces.extTag('Extension')]
        else:
            subElementNames.append(namespaces.extTag('Extension'))
        attrname = '_ext_ext'
        super().__init__(attrname, subElementNames, defaultPyValue)
        self._converter = None

    def __set__(self, instance, pyValue):
        if not isinstance(pyValue, ExtensionLocalValue):
            pyValue = ExtensionLocalValue(pyValue)
        super().__set__(instance, pyValue)

    def __get__(self, instance, owner):
        """Return a python value, uses the locally stored value."""
        if instance is None:  # if called via class
            return self
        try:
            value = getattr(instance, self._localVarName)
        except AttributeError:
            value = None
        if value is None:
            value = _PropertyValue(None, ExtensionLocalValue())
            setattr(instance, self._localVarName, value)
        return value.py_value

    def getPyValueFromNode(self, node):
        """Read value from node."""
        try:
            extension_node = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
        except ElementNotFoundException:
            return _PropertyValue(None, ExtensionLocalValue())
        return _PropertyValue(extension_node, ExtensionLocalValue(extension_node[:]))

    def updateXMLValue(self, instance, node):
        """Write value to node.

        The Extension Element is only added if there is at least one element available in local list.
        """
        try:
            property_value = getattr(instance, self._localVarName)
        except AttributeError:
            return  # nothing to add
        if property_value is None or len(property_value.py_value) == 0:
            return  # nothing to add
        sub_node = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
        sub_node.extend(copy_node_wo_parent(x) for x in property_value.py_value)


class SubElementProperty(_PropertyBase):
    """ uses a value that has an "asEtreeNode" method"""

    def __init__(self, subElementNames, valueClass, defaultPyValue=None, impliedPyValue=None):
        attrname = None
        super(SubElementProperty, self).__init__(attrname, subElementNames, defaultPyValue, impliedPyValue)
        self.valueClass = valueClass

    def getPyValueFromNode(self, node):
        value = self._defaultPyValue
        try:
            subNode = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
            value = self.valueClass.fromNode(subNode)
        except ElementNotFoundException:
            pass
        if value is None:
            return None
        return _PropertyValue(None, value)

    def updateXMLValue(self, instance, node):
        try:
            property_value = getattr(instance, self._localVarName)
            pyValue = property_value.py_value
        except AttributeError:
            pyValue = self._defaultPyValue

        if pyValue is not None:
            parentElement = self._getElementbyChildNamesList(node, self._subElementNames[:-1], createMissingNodes=True)
            self.rmLastSubElement(node)
            parentElement.append(pyValue.asEtreeNode(self._subElementNames[-1], parentElement.nsmap))


class SubElementListProperty(_ListPropertyBase):
    """ a list of values that have an "asEtreeNode" method. Used if maxOccurs="Unbounded" in BICEPS_ParticipantModel"""

    def __init__(self, subElementNames, cls):
        attrname = None
        super(SubElementListProperty, self).__init__(attrname, subElementNames)
        self._cls = cls

    def getPyValueFromNode(self, node):
        """ get from node"""
        objects = []
        try:
            pNode = self._getElementbyChildNamesList(node, self._subElementNames[:-1],
                                                     createMissingNodes=False)  # get parent Node
            nodes = pNode.findall(self._subElementNames[-1])
            for n in nodes:
                objects.append(self._cls.fromNode(n))
            return objects
        except ElementNotFoundException:
            return objects

    def updateXMLValue(self, instance, node):
        """ value is a list of objects with "asEtreeNode" method"""
        # remove all existing nodes
        try:
            pyValue = getattr(instance, self._localVarName)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            pyValue = self._defaultPyValue

        pNode = self._getElementbyChildNamesList(node, self._subElementNames[:-1],
                                                 createMissingNodes=True)  # get parent Node
        nodes = pNode.findall(self._subElementNames[-1])
        for n in nodes:
            pNode.remove(n)
        # ... and create new ones
        if pyValue is not None:
            for v in pyValue:
                pNode.append(v.asEtreeNode(self._subElementNames[-1], pNode.nsmap))

    def __str__(self):
        if self._subElementNames:
            path_string = ', '.join([str(x) for x in self._subElementNames])
            return '{} datatype {} in subelement {}'.format(self.__class__.__name__, self._cls.__name__, path_string)
        else:
            return '{} datatype {}'.format(self.__class__.__name__, self._cls.__name__)


class SubElementTextListProperty(_PropertyBase):
    """ represents a list of strings."""

    def __init__(self, subElementNames, noEmptySubNode=True):
        """

        :param subElementNames: path to the text elements
        :param noEmptySubNode: if true, the sub elements are not created if pyValue is None or an empty list.
                Otherwise the subelements except the last one are created.
        """
        attrname = None
        self._noEmptySubNode = noEmptySubNode
        super(SubElementTextListProperty, self).__init__(attrname, subElementNames, defaultPyValue=None)
        self._defaultPyValue = []

    def getPyValueFromNode(self, node):
        """ get from node"""
        objects = []
        try:
            pNode = self._getElementbyChildNamesList(node, self._subElementNames[:-1],
                                                     createMissingNodes=False)  # get parent Node
            nodes = pNode.findall(self._subElementNames[-1])
            for n in nodes:
                objects.append(n.text)
        except ElementNotFoundException:
            pass
        return _PropertyValue(None, objects)

    def updateXMLValue(self, instance, node):
        """ value is a list of strings"""
        # remove all existing nodes
        try:
            property_value = getattr(instance, self._localVarName)
            pyValue = property_value.py_value
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            pyValue = self._defaultPyValue

        if pyValue is None:
            return
        elif len(pyValue) == 0 and self._noEmptySubNode:
            return

        pNode = self._getElementbyChildNamesList(node, self._subElementNames[:-1],
                                                 createMissingNodes=True)  # get parent Node
        nodes = pNode.findall(self._subElementNames[-1])
        for n in nodes:
            pNode.remove(n)
        # ... and create new ones
        for v in pyValue:
            child = etree_.SubElement(pNode, self._subElementNames[-1])
            child.text = v

    def __str__(self):
        if self._subElementNames:
            path_string = ', '.join([str(x) for x in self._subElementNames])
            return '{} datatype {} in subelement {}'.format(self.__class__.__name__, self._cls.__name__, path_string)
        else:
            return '{} datatype {}'.format(self.__class__.__name__, self._cls.__name__)


class DateOfBirthProperty(_PropertyBase):
    """ this represents the DateOfBirth type of BICEPS schema:
        <xsd:simpleType>
            <xsd:union memberTypes="xsd:dateTime xsd:date xsd:gYearMonth xsd:gYear"/>
        </xsd:simpleType>
    xsd:dateTime is YYYY-MM-DDThh:mm:ss.sss
    xsd:date is YYYY-MM-DD format. All components are required
    xsd:gYearMonth is YYYY-MM
    xsd:gYear is YYYY
    If the timepoint of birth matters, the value SHALL be populated with a time zone.

    Time zone info can be provided:
       UTC can be specified by appending a Z character, e.g. 2002-09-24Z
       other timezones by adding a positive or negative time behind the date, e.g. 2002.09-24-06:00, 2002-09-24+06:00
    xsd:time is hh:mm:ss format, e.g. 9:30:10, 9:30:10.5. All components are required.
    Time zone handling is identical to date type

    The corresponding Python types are datetime.Date (=> not time point available) or datetime.Datetime (with time point attribute)
    """

    def __init__(self, subElementNames, defaultPyValue=None, impliedPyValue=None):
        attrname = None
        super().__init__(attrname, subElementNames, defaultPyValue, impliedPyValue)

    def getPyValueFromNode(self, node):
        try:
            node = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=False)
            if node is not None:
                date_string = node.text
                value = isoduration.parse_date_time(date_string)
                return _PropertyValue(date_string, value)
        except ElementNotFoundException:
            return None

    def updateXMLValue(self, instance, node):
        try:
            property_value = getattr(instance, self._localVarName)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            property_value = None

        if property_value is None:
            self.rmLastSubElement(node)
        else:
            date_string = None
            if property_value.xml_value is not None:
                date_string = property_value.xml_value  # use strings as they are
            elif property_value.py_value is not None:
                if isinstance(property_value.py_value, str):
                    date_string = property_value.py_value  # use strings as they are
                else:
                    date_string = self._mk_datestring(property_value.py_value)
            if date_string is not None:
                subElement = self._getElementbyChildNamesList(node, self._subElementNames, createMissingNodes=True)
                subElement.text = date_string

    @staticmethod
    def mk_value_object(date_string):
        return isoduration.parse_date_time(date_string)

    @staticmethod
    def _mk_datestring(date_object):
        return isoduration.date_time_string(date_object)


class TimestampAttributeProperty(NodeAttributeProperty):
    """ XML notation is integer in milliseconds.
    Python is a float in seconds."""

    def __init__(self, attrname, subElementNames=None, defaultPyValue=None, impliedPyValue=None):
        super(TimestampAttributeProperty, self).__init__(attrname, subElementNames, valueConverter=TimestampConverter,
                                                         defaultPyValue=defaultPyValue, impliedPyValue=impliedPyValue)


class CurrentTimestampAttributeProperty(NodeAttributeProperty):
    """ used for ClockState, it always writes current time to node. Setting value from python is possible, but makes no sense.
    """

    def __init__(self, attrname, subElementNames=None):
        super(CurrentTimestampAttributeProperty, self).__init__(attrname, subElementNames,
                                                                valueConverter=TimestampConverter, defaultPyValue=None)

    def updateXMLValue(self, instance, node):
        setattr(instance, self._localVarName, _PropertyValue(None, time.time()))
        super(CurrentTimestampAttributeProperty, self).updateXMLValue(instance, node)


class DecimalAttributeProperty(NodeAttributeProperty):
    """ XML notation is integer in milliseconds.
    Python is a float in seconds."""

    def __init__(self, attrname, subElementNames=None, defaultPyValue=None, impliedPyValue=None):
        super(DecimalAttributeProperty, self).__init__(attrname, subElementNames, valueConverter=DecimalConverter,
                                                       defaultPyValue=defaultPyValue, impliedPyValue=impliedPyValue)


class DurationAttributeProperty(NodeAttributeProperty):
    """ XML notation is integer in milliseconds.
    Python is a float in seconds."""

    def __init__(self, attrname, subElementNames=None, defaultPyValue=None, impliedPyValue=None):
        super(DurationAttributeProperty, self).__init__(attrname, subElementNames, valueConverter=DurationConverter,
                                                        defaultPyValue=defaultPyValue, impliedPyValue=impliedPyValue)


class IntegerAttributeProperty(NodeAttributeProperty):
    """ XML notation is integer in milliseconds.
    Python is a float in seconds."""

    def __init__(self, attrname, subElementNames=None, defaultPyValue=None, impliedPyValue=None):
        super(IntegerAttributeProperty, self).__init__(attrname, subElementNames, valueConverter=IntegerConverter,
                                                       defaultPyValue=defaultPyValue, impliedPyValue=impliedPyValue)


class BooleanAttributeProperty(NodeAttributeProperty):
    """ XML notation is integer in milliseconds.
    Python is a float in seconds."""

    def __init__(self, attrname, subElementNames=None, defaultPyValue=None, impliedPyValue=None):
        super(BooleanAttributeProperty, self).__init__(attrname, subElementNames, valueConverter=BooleanConverter,
                                                       defaultPyValue=defaultPyValue, impliedPyValue=impliedPyValue)


ZERO = datetime.timedelta(0)


class UTC(datetime.tzinfo):
    """Fixed offset in minutes east from UTC."""

    def __init__(self, offset_minutes, tzname=None):
        self._offset = datetime.timedelta(minutes=offset_minutes)
        self._tzname = tzname

    def utcoffset(self, dt):  # pylint:disable=unused-argument
        return self._offset

    def tzname(self, dt):  # pylint:disable=unused-argument
        return self._tzname

    def dst(self, dt):  # pylint:disable=unused-argument
        return ZERO



from .. import namespaces
from .. import sdcdevice
from .. pmtypes import ComponentActivation, MetricCategory, CodedValue
from . import providerbase


class GenericMetricProvider(providerbase.ProviderRole):
    """ Always added operations: None
    This is a generic Handler for
    - SetValueOperation on numeric metrics
    - SetStringOperation on (enum) string metrics
    """
    def __init__(self, activationStateCanRemoveMetricValue=True, log_prefix=None):
        """

        :param activationStateCanRemoveMetricValue: if True, SF717 is handled
               SF717: A Metric Provider shall not provide a MetricValue if the ActivationState = Shtdn|Off|Fail.
        """
        super().__init__(log_prefix)
        self.activationStateCanRemoveMetricValue = activationStateCanRemoveMetricValue

    def makeOperationInstance(self, operationDescriptorContainer):
        """ Can handle following cases:
        SetValueOperation, target = NumericMetricDescriptor: => handler = _setNumericValue
        SetStringOperation, target = (Enum)StringMetricDescriptor: => handler = _setString
        SetMetricStateOperationDescriptor, target = any subclass of AbstractMetricDescriptor: => handler = _setMetricState
        """
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        operationTargetDescriptorContainer = self._mdib.descriptions.handle.getOne(operationTargetHandle)

        if operationTargetDescriptorContainer.NODETYPE not in (namespaces.domTag('StringMetricDescriptor'),
                                                               namespaces.domTag('EnumStringMetricDescriptor'),
                                                               namespaces.domTag('NumericMetricDescriptor'),
                                                               namespaces.domTag('RealTimeSampleArrayMetricDescriptor')):
            return None # this is not metric provider role

        if operationDescriptorContainer.NODETYPE == namespaces.domTag('SetValueOperationDescriptor'):
            if operationTargetDescriptorContainer.NODETYPE == namespaces.domTag('NumericMetricDescriptor'):
                return self._mkOperation(sdcdevice.sco.SetValueOperation,
                                         handle=operationDescriptorContainer.handle,
                                         operationTargetHandle=operationTargetHandle,
                                         codedValue=operationDescriptorContainer.Type,
                                         currentArgumentHandler=self._setNumericValue)
            return None
        elif operationDescriptorContainer.NODETYPE == namespaces.domTag('SetStringOperationDescriptor'):
            if operationTargetDescriptorContainer.NODETYPE in (namespaces.domTag('StringMetricDescriptor'),
                                                               namespaces.domTag('EnumStringMetricDescriptor')):
                return self._mkOperation(sdcdevice.sco.SetStringOperation,
                                         handle=operationDescriptorContainer.handle,
                                         operationTargetHandle=operationTargetHandle,
                                         codedValue=operationDescriptorContainer.Type,
                                         currentArgumentHandler=self._setString)
            return None
        elif operationDescriptorContainer.NODETYPE == namespaces.domTag('SetMetricStateOperationDescriptor'):
            operation = self._mkOperation(sdcdevice.sco.SetMetricStateOperation,
                                          handle=operationDescriptorContainer.handle,
                                          operationTargetHandle=operationTargetHandle,
                                          codedValue=operationDescriptorContainer.Type,
                                          currentArgumentHandler=self._setMetricState)
            return operation
        return None

    def _setMetricState(self, operationInstance, value):
        """

        :param operationInstance: the operation
        :param value: a list of proposed metric states
        :return:
        """
        #ToDo: consider ModifiableDate attribute
        operationInstance.currentValue = value
        with self._mdib.mdibUpdateTransaction() as mgr:
            for proposedMetricState in value:
                state = mgr.getMetricState(proposedMetricState.descriptorHandle)
                if state.isMetricState:
                    self._logger.info('updating {} with proposed metric state', state)
                    state.updateFromOtherContainer(proposedMetricState, skippedProperties=['StateVersion', 'DescriptorVersion'])
                else:
                    self._logger.warn('_setMetricState operation: ignore invalid referenced type {} in operation', state.NODETYPE)

    def onPreCommit(self, mdib, transaction):
        if not self.activationStateCanRemoveMetricValue:
            return
        if transaction.metricStateUpdates:
            self._handleMetricsComponentActivation(transaction.metricStateUpdates.values())
        if transaction.rtSampleStateUpdates:
            self._handleMetricsComponentActivation(transaction.rtSampleStateUpdates.values())

    def _handleMetricsComponentActivation(self, metricStateUpdates):
        # check if MetricValue shall be removed
        for oldstate, newstate in metricStateUpdates:
            if newstate is None or not newstate.isMetricState:
                continue
            # SF717: check if MetricValue shall be automatically removed
            if newstate.ActivationState in (ComponentActivation.OFF,
                                            ComponentActivation.SHUTDOWN,
                                            ComponentActivation.FAILURE):
                if newstate.metricValue is not None:
                    # remove metric value
                    self._logger.info('{}: remove metric value because ActivationState="{}", handle="{}"',
                                      self.__class__.__name__, newstate.ActivationState, newstate.descriptorHandle)
                    newstate.metricValue = None



"""
This Module contains code handles Service Controller operations (sco).
All remote control commands of a client are executed by sco's

These operations share a common behavior:
A remote control command is executed async. The respone to such soap request contains a state (typically 'wait') and a transaction id.
The progress of the transaction is reported with an OperationInvokedReport.
A client must subscribe to the OperationInvokeReport Event of the 'Set' service, otherwise it would not get informed about progress.
"""
import queue
import threading
import time
import traceback
from collections import namedtuple

from lxml import etree as etree_

from sdc11073 import msgtypes
from .. import loghelper
from .. import mdib as mdib_
from .. import namespaces
from .. import observableproperties as properties
from .. import pmtypes
from ..mdib import msgreader


class _OperationsWorker(threading.Thread):
    """ Thread that enqueues and processes all operations.
    It manages transaction ids for all operations.
    Progress notifications are sent via subscription manager."""

    def __init__(self, operations_registry, subscriptionsmgr, mdib, log_prefix):
        """

        :param operations_registry:
        :param subscriptionsmgr: subscriptionsmgr.notifyOperation is called in order to notify all subscribers of OperationInvokeReport Events
        :param mdib:
        :param log_prefix:
        """
        super().__init__(name='DeviceOperationsWorker')
        self.daemon = True
        self._operations_registry = operations_registry
        self._subscriptionsmgr = subscriptionsmgr
        self._mdib = mdib
        self._operationsQ = queue.Queue(10)  # spooled operations
        self._transactionId = 1
        self._transactionIdLock = threading.Lock()
        self._logger = loghelper.getLoggerAdapter('sdc.device.op_worker', log_prefix)

    def enqueueOperation(self, operation, request):
        """ enqueues operation "operation".
        :param operation: a callable with signature operation(request, mdib)
        :param request: the soapEnvelope of the request
        @return: a transaction Id
        """
        with self._transactionIdLock:
            transactionId = self._transactionId
            self._transactionId += 1
        self._operationsQ.put((transactionId, operation, request), timeout=1)
        return transactionId

    def run(self):
        while True:
            try:
                try:
                    op = self._operationsQ.get(timeout=1.0)
                    tr_id, operation, request = op
                except queue.Empty:
                    self._operations_registry.check_invocation_timeouts()
                else:
                    if tr_id == 'stop':
                        self._logger.info('stop request found. Terminating now.')
                        return
                    time.sleep(0.001)
                    self._logger.info('{}: starting operation "{}"', operation.__class__.__name__, operation.handle)
                    # duplicate the WAIT respnse to the operation request as notification. Standard requires this.
                    self._subscriptionsmgr.notifyOperation(self._mdib.mdib_version_group, tr_id,
                                                           operation.handle, pmtypes.InvocationState.WAIT)
                    time.sleep(0.001)  # not really necessary, but in real world there might also be some delay.
                    self._subscriptionsmgr.notifyOperation(self._mdib.mdib_version_group, tr_id,
                                                           operation.handle, pmtypes.InvocationState.START)
                    try:
                        operation.executeOperation(request)
                        operation.last_called_time = time.time()

                        self._logger.info('{}: successfully finished operation "{}"', operation.__class__.__name__,
                                          operation.handle)
                        self._subscriptionsmgr.notifyOperation(self._mdib.mdib_version_group, tr_id,
                                                               operation.handle, pmtypes.InvocationState.FINISHED)
                    except Exception as ex:
                        self._logger.info('{}: error executing operation "{}": {}', operation.__class__.__name__,
                                          operation.handle, traceback.format_exc())
                        self._subscriptionsmgr.notifyOperation(self._mdib.mdib_version_group, tr_id,
                                                               operation.handle, pmtypes.InvocationState.FAILED,
                                                               error='Oth', errorMessage=repr(ex))
            except Exception as ex:
                self._logger.error('unexpected error while handling operation : {}', traceback.format_exc())

    def stop(self):
        self._operationsQ.put(('stop', None, None))  # a dummy request to stop the thread
        self.join(timeout=1)


class ScoOperationsRegistry(object):
    """ Registry for Sco operations.
    from BICEPS:
    A service control object to define remote control operations. Any pm:AbstractOperationDescriptor/@OperationTarget
    within this SCO SHALL only reference this or child descriptors within the CONTAINMENT TREE.
    NOTE - In modular systems, dynamically plugged-in modules would typically be modeled as VMDs.
    Such VMDs potentially have their own SCO. In every other case, SCO operations are modeled in pm:MdsDescriptor/pm:Sco.
    """

    def __init__(self, subscriptionsmgr, mdib, handle='_sco', log_prefix=None):
        self._worker = None
        self._subscriptionsmgr = subscriptionsmgr
        self._mdib = mdib
        self._log_prefix = log_prefix
        self._logger = loghelper.getLoggerAdapter('sdc.device.op_reg', log_prefix)
        self._registeredOperations = {}  # lookup by handle
        self._handle = handle

        # find the Sco of the Mds, this will be the default sco for new operations
        mdsDescriptorContainer = mdib.descriptions.NODETYPE.getOne(namespaces.domTag('MdsDescriptor'))
        scos = mdib.descriptions.find(parentHandle=mdsDescriptorContainer.handle).find(
            nodeName=namespaces.domTag('Sco')).objects
        if len(scos) == 1:
            self._logger.info('found Sco node in mds, using it')
            self._mds_sco_descriptorContainer = scos[0]
        else:
            self._logger.info('not found Sco node in mds, creating it')
            # create sco and add to mdib
            sco = etree_.Element(namespaces.domTag('Sco'),
                                 attrib={'DescriptorVersion': '1',
                                         namespaces.QN_TYPE: '{}:ScoDescriptor'.format(mdib.nsmapper.domPrefix()),
                                         'Handle': self._handle},
                                 nsmap=mdib.nsmapper.docNssmap)
            self._mds_sco_descriptorContainer = mdib_.descriptorcontainers.ScoDescriptorContainer.fromNode(
                mdib.nsmapper, sco, mdsDescriptorContainer.handle)

            mdib.descriptions.addObject(self._mds_sco_descriptorContainer)

    def check_invocation_timeouts(self):
        for op in self._registeredOperations.values():
            op.checkTimeout()

    def registerOperation(self, operation, scoDescriptorContainer=None):
        self._logger.info('register operation "{}"', operation)
        if operation.handle in self._registeredOperations:
            self._logger.info('handle {} is already registered, will re-use it', operation.handle)
        parentContainer = scoDescriptorContainer or self._mds_sco_descriptorContainer
        operation.setMdib(self._mdib, parentContainer)
        self._registeredOperations[operation.handle] = operation

    def unRegisterOperationByHandle(self, operationHandle):
        del self._registeredOperations[operationHandle]

    def getOperationByHandle(self, operationHandle):
        return self._registeredOperations.get(operationHandle)

    def enqueueOperation(self, operation, request):
        """ enqueues operation "operation".
        :param operation: a callable with signature operation(request, mdib)
        :param request: the soapEnvelope of the request
        @return: a transaction Id
        """
        return self._worker.enqueueOperation(operation, request)

    def startWorker(self):
        if self._worker is not None:
            raise RuntimeError('SCO worker is already running')
        self._worker = _OperationsWorker(self, self._subscriptionsmgr, self._mdib, self._log_prefix)
        self._worker.start()

    def stopWorker(self):
        if self._worker is not None:
            self._worker.stop()
            self._worker = None


class OperationDefinition(object):
    """ This is the base class of all provided operations.
    An operation is a point for remote control over the network."""
    currentValue = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    currentRequest = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    currentArgument = properties.ObservableProperty(fireOnlyOnChangedValue=False)
    onTimeout = properties.ObservableProperty(fireOnlyOnChangedValue=False)

    def __init__(self, handle, operationTarget,
                 operationDescriptorQName,
                 operationStateQName,
                 safetyClassification=pmtypes.SafetyClassification.INF,
                 codedValue=None,
                 safetyReq=None,
                 log_prefix=None,
                 extExtension=None):  # pylint:disable=too-many-arguments
        """
        :param handle: the handle of the operation itself.
        :param operationTarget: the handle of the modified data (MdDescription)
        :param operationDescriptorQName: the QName of the descriptor for this operation
        :param operationStateQName: the QName of the state for this operation
        :param safetyClassification: one of pmtypes.SafetyClassification values
        :param codedValue: a pmtypes.CodedValue instance
        :param safetyReq: a pmtypes.T_SafetyReq instance or None
        :param extExtension: list of etree.Element elements to be added to the Extension element
        """
        self._logger = loghelper.getLoggerAdapter('sdc.device.op.{}'.format(self.__class__.__name__), log_prefix)
        self._mdib = None
        self._descriptorContainer = None
        self._operationStateContainer = None
        self._operationTargetContainer = None
        self._handle = handle
        self._operationTargetHandle = operationTarget
        # documentation of operationTarget:
        # A HANDLE reference this operation is targeted to. In case of a single state this is the HANDLE of the descriptor.
        # In case that multiple states may belong to one descriptor (pm:AbstractMultiState), OperationTarget is the HANDLE
        # of one of the state instances (if the state is modified by the operation).
        self._operationDescriptorQName = operationDescriptorQName
        self._operationStateQName = operationStateQName

        self._safetyClassification = safetyClassification
        self._codedValue = codedValue
        self._extExtension = extExtension
        self.safetyReq = safetyReq  # can become a pmtypes.T_SafetyReq instance
        self.calls = []  # record when operation was called
        self.last_called_time = None

    @property
    def handle(self):
        return self._handle

    @property
    def operationTarget(self):
        return self._operationTargetHandle

    @property
    def operationTargetStorage(self):
        return self._mdib.states

    def executeOperation(self, request):  # pylint: disable=unused-argument
        """ This is the code that executes the operation itself.
        A handler that executes the operation must be bound to observable "currentRequest"."""
        self.calls.append((time.time(), request))
        self.currentRequest = request

    def checkTimeout(self):
        if self.last_called_time is None:
            return
        if self._descriptorContainer.InvocationEffectiveTimeout is None:
            return
        age = time.time() - self.last_called_time
        if age < self._descriptorContainer.InvocationEffectiveTimeout:
            return
        self.onTimeout = True  # let observable fire
        self.last_called_time = None  # reset time

    def setMdib(self, mdib, parentDescriptorContainer):
        """ The operation needs to know the mdib that it operates on.
        This is called by SubscriptionManager on registration.
        Needs to be implemented by derived classes if specific things have to be initialized."""
        if self._mdib is not None:
            raise RuntimeError('Mdib is already set')
        self._mdib = mdib
        self._logger.log_prefix = mdib.log_prefix  # use same prefix as mdib for logging
        self._descriptorContainer = self._mdib.descriptions.handle.getOne(self._handle, allowNone=True)
        if self._descriptorContainer is not None:
            # there is already a descriptor
            self._logger.info('descriptor for operation "{}" is already present, re-using it'.format(self._handle))
        else:
            operationDescriptorClass = mdib.getDescriptorContainerClass(self._operationDescriptorQName)
            self._descriptorContainer = operationDescriptorClass(mdib.nsmapper,
                                                                 namespaces.domTag('Operation'),
                                                                 self._handle, parentDescriptorContainer.handle,
                                                                 )
            self._initOperationDescriptorContainer()
            self._descriptorContainer.updateNode()
            mdib.descriptions.addObject(self._descriptorContainer)

        self._operationStateContainer = self._mdib.states.descriptorHandle.getOne(self._handle, allowNone=True)
        if self._operationStateContainer is not None:
            self._logger.info('operation state for operation "{}" is already present, re-using it'.format(self._handle))
            self._operationStateContainer.updateNode()
        else:
            operationStateClass = mdib.getStateContainerClass(self._operationStateQName)
            self._operationStateContainer = operationStateClass(mdib.nsmapper, self._descriptorContainer)
            self._operationStateContainer.updateNode()
            mdib.states.addObject(self._operationStateContainer)

        # now add the object that is target of operation
        self._initOperationTargetContainer()

    def _initOperationDescriptorContainer(self):
        self._descriptorContainer.SafetyReq = self.safetyReq
        self._descriptorContainer.OperationTarget = self._operationTargetHandle
        self._descriptorContainer.SafetyClassification = self._safetyClassification
        if self._codedValue is not None:
            self._descriptorContainer.Type = self._codedValue
        if self._extExtension:
            self._descriptorContainer.ext_Extension = etree_.Element(namespaces.extTag('Extension'))
            for node in self._extExtension:
                self._descriptorContainer.ext_Extension.append(node)

    def _initOperationTargetContainer(self):
        """ Create the object that is manipulated by the operation"""
        operationTargetDescriptor = self._mdib.descriptions.handle.getOne(self._operationTargetHandle)
        self._operationTargetContainer = self._mdib.states.descriptorHandle.getOne(self._operationTargetHandle,
                                                                                   allowNone=True)  # pylint:disable=protected-access
        if self._operationTargetContainer is not None:
            self._logger.info('operation target state for operation "{}" is already present, re-using it'.format(
                self._operationTargetHandle))
        else:
            self._operationTargetContainer = self._mdib.mkStateContainerFromDescriptor(
                operationTargetDescriptor)  # pylint:disable=protected-access
            self._operationTargetContainer.updateNode()
            self._logger.info('creating {} DescriptorHandle = {}', self._operationTargetContainer.__class__.__name__,
                              self._operationTargetHandle)
            if self._operationTargetContainer is not None:
                self.operationTargetStorage.addObject(self._operationTargetContainer)

    def setOperatingMode(self, mode):
        """ Mode is one of En, Dis, NA"""
        with self._mdib.mdibUpdateTransaction() as tr:
            st = tr.getOperationalState(self._handle)
            st.OperatingMode = mode

    def collectValues(self, numberOfValues=None):
        """ Async way to retrieve next value(s):
        Returns a Future-like object that has a result() method.
        For details see properties.SingleValueCollector and propertiesValuesCollector documentation.
        """
        if numberOfValues is None:
            return properties.SingleValueCollector(self, 'currentValue')
        else:
            return properties.ValuesCollector(self, 'currentValue', numberOfValues)

    def __str__(self):
        return '{} handle={} operationTarget={}'.format(self.__class__.__name__, self._handle,
                                                        self._operationTargetHandle)


class SetStringOperation(OperationDefinition):
    OP_DESCR_QNAME = namespaces.domTag('SetStringOperationDescriptor')
    OP_STATE_QNAME = namespaces.domTag('SetStringOperationState')

    def __init__(self, handle, operationTarget, initialValue=None, codedValue=None, extExtension=None):
        super(SetStringOperation, self).__init__(
            handle=handle,
            operationTarget=operationTarget,
            operationDescriptorQName=namespaces.domTag('SetStringOperationDescriptor'),
            operationStateQName=namespaces.domTag('SetStringOperationState'),
            codedValue=codedValue,
            extExtension=extExtension)
        self.currentValue = initialValue

    def executeOperation(self, request):  # pylint: disable=unused-argument
        super(SetStringOperation, self).executeOperation(request)
        stringNode = request.bodyNode.xpath('*/msg:RequestedStringValue', namespaces=namespaces.nsmap)
        if stringNode:
            self.currentArgument = str(stringNode[0].text)

    @classmethod
    def fromOperationContainer(cls, operationContainer):
        return cls(handle=operationContainer.handle,
                   operationTarget=operationContainer.OperationTarget,
                   initialValue=None, codedValue=None)


class SetValueOperation(OperationDefinition):
    OP_DESCR_QNAME = namespaces.domTag('SetValueOperationDescriptor')
    OP_STATE_QNAME = namespaces.domTag('SetValueOperationState')

    def __init__(self, handle, operationTarget, initialValue=None, codedValue=None, extExtension=None):
        super(SetValueOperation, self).__init__(
            handle=handle,
            operationTarget=operationTarget,
            operationDescriptorQName=namespaces.domTag('SetValueOperationDescriptor'),
            operationStateQName=namespaces.domTag('SetValueOperationState'),
            codedValue=codedValue,
            extExtension=extExtension)
        self.currentValue = initialValue

    def executeOperation(self, request):  # pylint: disable=unused-argument
        super(SetValueOperation, self).executeOperation(request)
        valueNodes = request.bodyNode.xpath('*/msg:RequestedNumericValue', namespaces=namespaces.nsmap)
        if valueNodes:
            self.currentArgument = float(valueNodes[0].text)


class SetContextStateOperation(OperationDefinition):
    """Default implementation of SetContextOperation."""
    OP_DESCR_QNAME = namespaces.domTag('SetContextStateOperationDescriptor')
    OP_STATE_QNAME = namespaces.domTag('SetContextStateOperationState')

    def __init__(self, handle, operationTarget, codedValue=None, extExtension=None):
        super(SetContextStateOperation, self).__init__(
            handle,
            operationTarget,
            operationDescriptorQName=namespaces.domTag('SetContextStateOperationDescriptor'),
            operationStateQName=namespaces.domTag('SetContextStateOperationState'),
            codedValue=codedValue,
            extExtension=extExtension)

    def executeOperation(self, request):  # pylint: disable=unused-argument
        super(SetContextStateOperation, self).executeOperation(request)
        # write currentArgument observable
        proposedContextStateNodes = request.bodyNode.xpath('*/msg:ProposedContextState', namespaces=namespaces.nsmap)
        msg_reader = msgreader.MessageReader(self._mdib)
        self.currentArgument = [msg_reader.mkStateContainerFromNode(p) for p in proposedContextStateNodes]

    @property
    def operationTargetStorage(self):
        return self._mdib.contextStates

    def _initOperationTargetContainer(self):
        """ initially no patient context is created."""
        pass

    @classmethod
    def fromOperationContainer(cls, operationContainer):
        return cls(handle=operationContainer.handle,
                   operationTarget=operationContainer.OperationTarget)


RecordedCall = namedtuple('RecordedCall', 'timestamp args')


class ActivateOperation(OperationDefinition):
    """ This default implementation only registers calls, no manipulation of operation target
    """
    OP_DESCR_QNAME = namespaces.domTag('ActivateOperationDescriptor')
    OP_STATE_QNAME = namespaces.domTag('ActivateOperationState')

    def __init__(self, handle, operationTarget, codedValue=None, safetyReq=None, extExtension=None):
        super(ActivateOperation, self).__init__(
            handle=handle,
            operationTarget=operationTarget,
            operationDescriptorQName=namespaces.domTag('ActivateOperationDescriptor'),
            operationStateQName=namespaces.domTag('ActivateOperationState'),
            codedValue=codedValue,
            safetyReq=safetyReq,
            extExtension=extExtension)

    def executeOperation(self, request):  # pylint: disable=unused-argument
        super(ActivateOperation, self).executeOperation(request)
        self.currentArgument = request.bodyNode.xpath('*/msg:Argument/msg:ArgValue/text()', namespaces=namespaces.nsmap)


class SetAlertStateOperation(OperationDefinition):
    """ This default implementation only registers calls, no manipulation of operation target
    """
    OP_DESCR_QNAME = namespaces.domTag('SetAlertStateOperationDescriptor')
    OP_STATE_QNAME = namespaces.domTag('SetAlertStateOperationState')

    def __init__(self, handle, operationTarget, codedValue=None, safetyReq=None, log_prefix=None, extExtension=None):
        super(SetAlertStateOperation, self).__init__(
            handle=handle,
            operationTarget=operationTarget,
            operationDescriptorQName=namespaces.domTag('SetAlertStateOperationDescriptor'),
            operationStateQName=namespaces.domTag('SetAlertStateOperationState'),
            codedValue=codedValue,
            safetyReq=safetyReq,
            log_prefix=log_prefix,
            extExtension=extExtension)

    def executeOperation(self, request):
        super(SetAlertStateOperation, self).executeOperation(request)
        proposedAlertStateNodes = request.bodyNode.xpath('*/msg:ProposedAlertState', namespaces=namespaces.nsmap)
        if len(proposedAlertStateNodes) > 1:  # schema allows exactly one ProposedAlertState:
            raise ValueError(
                'only one ProposedAlertState argument allowed, found {}'.format(len(proposedAlertStateNodes)))
        if len(proposedAlertStateNodes) == 0:
            raise ValueError('no ProposedAlertState argument found')
        else:
            msg_reader = msgreader.MessageReader(self._mdib)
            self.currentArgument = msg_reader.mkStateContainerFromNode(proposedAlertStateNodes[0])


class SetComponentStateOperation(OperationDefinition):
    """ This default implementation only registers calls, no manipulation of operation target
    """
    OP_DESCR_QNAME = namespaces.domTag('SetComponentStateOperationDescriptor')
    OP_STATE_QNAME = namespaces.domTag('SetComponentStateOperationState')

    def __init__(self, handle, operationTarget, codedValue=None, safetyReq=None, log_prefix=None, extExtension=None):
        super(SetComponentStateOperation, self).__init__(
            handle=handle,
            operationTarget=operationTarget,
            operationDescriptorQName=namespaces.domTag('SetComponentStateOperationDescriptor'),
            operationStateQName=namespaces.domTag('SetComponentStateOperationState'),
            codedValue=codedValue,
            safetyReq=safetyReq,
            log_prefix=log_prefix,
            extExtension=extExtension)

    def executeOperation(self, request):
        super(SetComponentStateOperation, self).executeOperation(request)
        proposedComponentStateNodes = request.bodyNode.xpath('*/msg:ProposedComponentState',
                                                             namespaces=namespaces.nsmap)
        msg_reader = msgreader.MessageReader(self._mdib)
        self.currentArgument = [msg_reader.mkStateContainerFromNode(p) for p in proposedComponentStateNodes]


class SetMetricStateOperation(OperationDefinition):
    """ This default implementation only registers calls, no manipulation of operation target
    """
    OP_DESCR_QNAME = namespaces.domTag('SetMetricStateOperationDescriptor')
    OP_STATE_QNAME = namespaces.domTag('SetMetricStateOperationState')

    def __init__(self, handle, operationTarget, codedValue=None, safetyReq=None, log_prefix=None, extExtension=None):
        super(SetMetricStateOperation, self).__init__(
            handle=handle,
            operationTarget=operationTarget,
            operationDescriptorQName=namespaces.domTag('SetMetricStateOperationDescriptor'),
            operationStateQName=namespaces.domTag('SetMetricStateOperationState'),
            codedValue=codedValue,
            safetyReq=safetyReq,
            log_prefix=log_prefix,
            extExtension=extExtension)

    def executeOperation(self, request):
        super(SetMetricStateOperation, self).executeOperation(request)
        proposedMetricStateNodes = request.bodyNode.xpath('*/msg:ProposedMetricState', namespaces=namespaces.nsmap)
        msg_reader = msgreader.MessageReader(self._mdib)
        self.currentArgument = [msg_reader.mkStateContainerFromNode(m) for m in proposedMetricStateNodes]


def getOperationClass(qname):
    if qname == namespaces.domTag('SetStringOperationDescriptor'):
        return SetStringOperation
    elif qname == namespaces.domTag('SetValueOperationDescriptor'):
        return SetValueOperation
    elif qname == namespaces.domTag('SetContextStateOperationDescriptor'):
        return SetContextStateOperation
    elif qname == namespaces.domTag('ActivateOperationDescriptor'):
        return ActivateOperation
    elif qname == namespaces.domTag('SetAlertStateOperationDescriptor'):
        return SetAlertStateOperation
    elif qname == namespaces.domTag('SetMetricStateOperationDescriptor'):
        return SetMetricStateOperation
    elif qname == namespaces.domTag('SetComponentStateOperationDescriptor'):
        return SetComponentStateOperation
    elif qname == namespaces.domTag('SetComponentStateOperationDescriptor'):
        return SetComponentStateOperation



"""
A minimal set of codes, used by role provider.
"""

class NomenclatureCodes(object):
    # only a small subset of all codes, these are used for tests
    MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE = 128284 # An operation to initiate global all audio pause
    MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE = 128285 # An operation to cancel global all audio pause
    MDC_OP_SET_TIME_SYNC_REF_SRC = 128505 # An operation to Set the active reference source of a clock for time synchronization.
    OP_SET_NTP = 194041  # deprecated REF_ID use MDC_OP_SET_TIME_SYNC_REF_SRC, needed for backwards compatibility
    OP_SET_TZ = 194040  # deprecated REF_ID use MDC_ACT_SET_TIME_ZONE, needed for backwards compatibility
    MDC_ACT_SET_TIME_ZONE = 68632 # replaces private Code MDC_OP_SET_TIME_ZONE (194040)
